/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * Extensible Shell example: Command Module.
 * This module adds the "load" command to the esh.
 */
#include <sys/types.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <pld.h>		/* Peludo PLD API */
#include <esh.h>		/* Extensible Shell API */

#define BUFFER_SIZE		4096


#define PRINT	0


/**
 * Read a PLD from a file descriptor.
 */
static PLDFILE *
receive(int fd)
{
	size_t size, remaining, pos;
	void *buf, *tmp;
	ssize_t r;
	
	size = 0;
	remaining = 0;
	pos = 0;
	buf = NULL;

	while (1) {
		/*
		 * If necessary expand the buffer to contain the PLD.
		 */
		if (remaining == 0) {
			tmp = realloc(buf, size + BUFFER_SIZE);
			if (tmp == NULL) {
				break;
			}
			remaining = BUFFER_SIZE;
			size += BUFFER_SIZE;
			buf = tmp;
		}

		/*
		 * Read PLD data.
		 */
		r = read(fd, buf + pos, remaining);
		if (r < 0) {
			break;
		}

		if (r == 0) {
			/*
			 * PLD module successfuly received.
			 */
			return ((PLDFILE *)buf);
		}

		remaining -= r;
		pos += r;
	}

	/*
	 * Reception failed.
	 */
	if (buf != NULL) {
		free(buf);
	}

	return NULL;
}


/**
 * "load" command.
 */
static void
load(int fd, char *args)
{
	PLDFILE *pld;
	PLSTATUS st;
	void *mod;
#if PRINT
	FILE *f;

	/*
	 * We can use the easier but slower fprintf(3) instad of write(2).
	 */
	f = fdopen(fd, "r+");
	if (f == NULL) {
		char msg[256];
		int len;

		len = snprintf(msg, sizeof(msg), "load: Can't fdopen: %s\n\r",
		    strerror(errno));
		write(fd, msg, len);
		return;
	}
#endif

	/*
	 * Read PLD from the file descriptor.
	 */
	pld = receive(fd);
	if (pld == NULL) {
#if PRINT
		fprintf(f, "load: Error receiving module: %s\n\r",
		    strerror(errno));
#endif
	} else {
		/*
		 * Install PLD.
		 */
		st = pld_open(&mod, pld);
		if (st == PLSTATUS_SUCCESS) {
#if PRINT
			fprintf(f, "load: New module installed. Handle: %p\n\r",
			    mod);
#endif
		} else {
#if PRINT
			fprintf(f, "load: Error installing module: %s\n\r",
			    strerror(errno));
#endif
			free(pld);
		}
	}
}


/**
 * Add the "load" command to the esh server.
 */
void
_init()
{
	esh_add("load",
	    "load\\n<data>\n\r"
	    "\tLoad a PLD module. You should use this command like this:\n\r"
	    "\t    $ (echo load; cat <module.pld>) | nc [-q0] server 4321\n\r"
	    "\tThe new module is added as soon as the connection is "
	    "closed.\n\r", load);
}


