/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * Extensible Shell example: Main Module.
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <pthread.h>

#include <plist.h>		/* Peludo List API		*/
#include <plversion.h>		/* Peludo version macros	*/

#include <esh.h>		/* Extensible Shell API		*/

#ifndef ESH_MAXCONN
#define ESH_MAXCONN		10
#endif

#ifndef ESH_PROMPT
#define ESH_PROMPT		"esh> "
#endif

#ifndef ESH_PORT
#define ESH_PORT		4321
#endif


/**
 * A registered command.
 */
typedef struct _ESHCMD {
	/**
	 * Next command.
	 */
	PLIST cmdnext;		/* First member, always */

	/**
	 * Command name.
	 */
	char name[10];

	/**
	 * Command help.
	 */
	char help[256];

	/**
	 * Command function.
	 */
	void (*command)(int, char *);

} ESHCMD;


/**
 * List of registered commands.
 * @see cmdlist_mutex
 */
static PLIST cmdlist;


/**
 * Mutex for the list of commands.
 * @see cmdlist
 */
static pthread_mutex_t cmdlist_mutex = PTHREAD_MUTEX_INITIALIZER;


/**
 * A client connetion parameters.
 */
typedef struct _CLIENTARGS {
	int fd;
	struct sockaddr_in addr;
} CLIENTARGS;


/**
 * List of client connections.
 * @see carg_mutex
 */
static CLIENTARGS carg[ESH_MAXCONN];


/**
 * Mutex for the list of client connections.
 * @see carg
 */
static pthread_mutex_t carg_mutex = PTHREAD_MUTEX_INITIALIZER;


/**
 * Print the greetings message into a buffer.
 * @param buf Buffer;
 * @param bufsiz Buffer size.
 * @return The length of the greeting message.
 */
static int
greetings(char *buf, int bufsiz)
{
	return snprintf(buf, bufsiz,
	    "Extensible Injectable Shell v1.0.0\n\r"
	    "Built using Peludo %s\n\r\n\r", PLRELEASE);
}


/**
 * "help" command.
 */
static void
cmd_help(int fd, char *args)
{
	ESHCMD *eshcmd;
	char msg[1024];
	int len;

	len = greetings(msg, sizeof(msg));
	write(fd, msg, len);

	pthread_mutex_lock(&cmdlist_mutex);

	PLIST_FOREACH(&cmdlist, eshcmd, cmdnext) {
		len = snprintf(msg, sizeof(msg), "%s\n\r", eshcmd->help);
		write(fd, msg, len);
	}

	pthread_mutex_unlock(&cmdlist_mutex);
}


/**
 * Read a line from a specified file descriptor.
 * @param fd File descritor;
 * @param buf Buffer to be filled with the input data;
 * @param buflen Size of the buffer.
 * @return The number of read bytes.
 */
static int
readline(int fd, char *buf, int buflen)
{
	int r, len;

	r = 0;

	while (buflen--) {
		len = read(fd, buf, 1);
		if (len <= 0) {
			if (len == 0 || errno == ECONNRESET) {
				break;
			}

			fprintf(stderr, "Can't read: %s\n", strerror(errno));
			exit(-1);
			/* NOTREACHED */
		}

		r++;
		buf++;

		if (*(buf - 1) == '\n') {
			break;
		}

	}

	*buf = '\0';

	return r;
}


/**
 * Process a given command.
 * @param fd Client's file descriptor;
 * @param cmd Command string;
 * @param args Command arguments.
 */
static void
process_command(int fd, char *cmd, char *args)
{
	ESHCMD *eshcmd;
	char msg[256];
	int len;

	pthread_mutex_lock(&cmdlist_mutex);

	PLIST_FOREACH(&cmdlist, eshcmd, cmdnext) {
		if (!strcasecmp(eshcmd->name, cmd)) {
			pthread_mutex_unlock(&cmdlist_mutex);
			eshcmd->command(fd, args);
			return;
		}
	}

	pthread_mutex_unlock(&cmdlist_mutex);

	len = snprintf(msg, sizeof(msg), "Command not found: %s\n\r", cmd);
	write(fd, msg, len);
}


/**
 * Handle a client connection.
 * @param _arg Client parameters.
 * @see CLIENTARGS
 */
static void *
process_client(void *_arg)
{
	char buf[1024], *cmd, *cmdargs;
	CLIENTARGS *arg;
	int len;

	arg = (CLIENTARGS *)_arg;

	/*
	 * Greetings.
	 */
	len = greetings(buf, sizeof(buf));
	write(arg->fd, buf, len);
	
	/*
	 * Command loop.
	 */
	while (1) {
		/*
		 * Shell prompt and read input line.
		 */
		write(arg->fd, ESH_PROMPT, strlen(ESH_PROMPT));
		len = readline(arg->fd, buf, sizeof(buf));
		if (len == 0 || !strcasecmp(buf, "quit\n")) {
			/*
			 * "quit" is the only hardcoded command.
			 */
			break;
		}

		/*
		 * Separate command from its argments.
		 */
		cmdargs = buf;
		cmd = strsep(&cmdargs, " \n\r\v\t");

		/*
		 * Remove leading and tailing blanks from the arguments.
		 */
		while (isspace(*cmdargs)) cmdargs++;
		if ((len = strlen(cmdargs)) != 0) {
			while (isspace(*(cmdargs + len - 1))) len--;
		}
		*(cmdargs + len) = '\0';

		/*
		 * Process command.
		 */
		if (*cmd != '\0') {
			process_command(arg->fd, cmd, cmdargs);
		}
	}

	close(arg->fd);

	pthread_mutex_lock(&carg_mutex);
	arg->fd = -1;
	pthread_mutex_unlock(&carg_mutex);

	return NULL;
}


/*
 * Extensible Shell API - Add a new command.
 * @param cmd Command name;
 * @param help Help string;
 * @param fn Command callback.
 * @return A non-null value on success; 0 on error.
 */
int PLEXPORT
esh_add(char *cmd, char *help, void (*fn)(int, char *))
{
	ESHCMD *eshcmd;
	int err = 0;

	pthread_mutex_lock(&cmdlist_mutex);

	PLIST_FOREACH(&cmdlist, eshcmd, cmdnext) {
		if (!strcasecmp(cmd, eshcmd->name)) {
			err = EEXIST;
			break;
		}
	}

	if (err == 0) {
		eshcmd = (ESHCMD *)malloc(sizeof(ESHCMD));
		if (eshcmd != NULL) {
			PLIST_INIT(&eshcmd->cmdnext);
			strlcpy(eshcmd->name, cmd, sizeof(eshcmd->name));
			strlcpy(eshcmd->help, help, sizeof(eshcmd->help));
			eshcmd->command = fn;
			plist_add_tail(&cmdlist, (PLIST *)eshcmd);
		} else {
			err = errno;
		}
	}

	pthread_mutex_unlock(&cmdlist_mutex);

	return err;
}


/**
 * Initialization function.
 * Initialize internal shell parameters and add the "help" command.
 *
 * <p>The Peludo Kernel calls this function during module initialization.</p>
 * <p>Think of this function as the standard unix _init() or the windows'
 * DllMain().</p>
 *
 * @see cmd_help(int, char *)
 */
void
_init()
{
	int i;

	/*
	 * Internal initialzation.
	 */
	PLIST_INIT(&cmdlist);
	for (i = 0; i < ESH_MAXCONN; i++) {
		carg[i].fd = -1;
	}

	/*
	 * Add the "help" command.
	 */
	esh_add("help", "help\n\r\tPrint usage information\n\r", cmd_help);
}


/**
 * Start the Exensible Shell server.
 */
int
main(int argc, char **argv)
{
	struct sockaddr_in addr, caddr;
	struct sigaction act;
	socklen_t caddr_len;
	int lfd, cfd, i;
	pthread_t ct;

	/*
	 * Prepare listening shell server.
	 */
	bzero(&act, sizeof(act));
	act.sa_handler = SIG_IGN;
	act.sa_flags = SA_RESTART;
	sigemptyset(&act.sa_mask);
	if (sigaction(SIGPIPE, &act, NULL) < 0) {
		fprintf(stderr, "Can't sigaction: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	lfd = socket(AF_INET, SOCK_STREAM, 0);
	if (lfd < 0) {
		fprintf(stderr, "Can't socket: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(ESH_PORT);

	i = SO_REUSEADDR;
	if (setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, (void *)&i, sizeof(i))
	    < 0) {
		fprintf(stderr, "Can't setsockopt: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	if (bind(lfd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		fprintf(stderr, "Can't bind: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	if (listen(lfd, 5) < 0) {
		fprintf(stderr, "Can't listen: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	/*
	 * Accept new connections.
	 */
	while (1) {
		caddr_len = sizeof(caddr);
		cfd = accept(lfd, (struct sockaddr *)&caddr, &caddr_len);
		if (cfd < 0) {
			if (errno == EINTR) {
				continue;
			}

			fprintf(stderr, "Can't accept: %s\n", strerror(errno));
			exit(-1);
			/* NOTREACHED */
		}

		/*
		 * Check if can handle a new connection.
		 */
		pthread_mutex_lock(&carg_mutex);
		for (i = 0; i < ESH_MAXCONN; i++) {
			if (carg[i].fd == -1) {
				carg[i].fd = cfd;
				carg[i].addr = caddr;
				break;
			}
		}
		pthread_mutex_unlock(&carg_mutex);

		if (i == ESH_MAXCONN) {
			static char *buf =
			    "Shell too busy. Try again later\n\r\n\r";
			write(cfd, buf, strlen(buf));
			close(cfd);
			continue;
		}

		/*
		 * Create a new thread to handle the new connection.
		 */
		i = pthread_create(&ct, NULL, (void *)process_client, &carg[i]);
		if (i != 0) {
			fprintf(stderr, "Can't pthread_create: %s\n",
			    strerror(i));
			exit (-1);
			/* NOTREACHED */
		}
	}
}

