/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/core/kernel/kinit.c
 * Peludo Kernel Initialization.
 */
#include <plc.h>
#include <pld.h>
#include <plzma.h>
#include <pllog.h>
#include <plversion.h>

#include "kernel.h"
#include "pldnames.h"

static PLSTATUS _kinit_getargs(long, int *, char ***);
static void _kinit_freeargs(int, char **);


/**
 * Peludo Kernel Entry Point.
 * <p>Initialize the peludo kernel, load the attached PLD Chain and call
 * the <code>int plstart(int argc, char **argv)</code> method of the first
 * module exporting it.</p>
 * @param argc Number of arguments in the command line;
 * @param argv Command line arguments;
 * @param pld Memory address containing the PLD Chain to load.
 *
 * <p><b>This function *must* be located at offset 0 of the code section.</b>
 * </p>
 *
 * @return A peludo status code or the Application's exit value.
 *
 * @see PLDFILE
 * @see plinit()
 * @see pld_init()
 * @see plzma_init()
 * @see pld_open_all()
 * @see _plstart()
 */
long __attribute__((section(".kinit")))
kinit(int argc, char **argv, long pld)
{
	long (*_plstart)(int, char **);
	long status, args;
	char **eargv;
	int eargc;
	PMOD *mod;

	if (pld == 0) {
		/*
 		 * No modules attached.
 		 */
		return PLSTATUS_SUCCESS;
	}

	/* Cannot use the PLC before initializing it */
	if ((status = plinit()) != PLSTATUS_SUCCESS) {
		return status;
	}

	pldebug("%s: Initializing Peludo Kernel %s build %s, base: %x, "
	    "kinit: %x\n",
	    __FUNCTION__,
	    PLRELEASE,
	    PLBUILD,
	    (long)kbase & (~PLPAGE_MASK),
	    kbase);

	/*
	 * Initialize kernel components.
	 */
	if ((status = plzma_init()) != PLSTATUS_SUCCESS) {
		pldebug("%s: Initialization of lzma decoder failed: "
		    "%d \"%s\"\n",
		    __FUNCTION__,
		    status,
		    plstrstatus(status));
		goto bye;
	}

	if ((status = pld_init(&pld)) != PLSTATUS_SUCCESS) {
		pldebug("%s: Initialization of PLD loader failed: %d \"%s\"\n",
		    __FUNCTION__,
		    status,
		    plstrstatus(status));
		goto bye;
	}

	/*
	 * Load attached modules.
	 */
	if ((status = pld_open_all((PLDFILE *)pld)) == PLSTATUS_SUCCESS) {
		/*
		 * If no options are specified use the embedded command line.
		 */
		eargc = 0;
		eargv = NULL;
		if (argc <= 1) {
			mod = pld_lookup(PLDNAME_FILE_CMDLINE);
			if (mod != NULL) {
				args = (long)pld_getdata(mod, NULL);
				status = _kinit_getargs(args, &eargc, &eargv);
				if (status != PLSTATUS_SUCCESS) {
					goto bye;
				}
				if (eargc > 0) {
					argc = eargc;
					argv = eargv;
				}
			}
		}

		/*
		 * Call _plstart()
		 */
		if (pld_sym_any((void **)&_plstart, (void **)&mod,
		    PLDNAME_SYMBOL__PLSTART) == PLSTATUS_SUCCESS) {

			pldebug("%s: Calling _plstart()...\n", __FUNCTION__);

#ifdef PLDEBUG
			PLTRAP();
#endif
			status = _plstart(argc, argv);

			pldebug("%s: _plstart() returned code: %d\n",
			    __FUNCTION__,
			    status);

		} else {
			pldebug("%s: _plstart() function not found.\n",
			    __FUNCTION__);
		}

		_kinit_freeargs(eargc, eargv);
		pld_release();

	} else {
		pldebug("%s: Module Initialization Failed: %d \"%s\"\n",
		    __FUNCTION__,
		    status,
		    plstrstatus(status));
	}

bye:	plrelease();

	return status;
}


static PLSTATUS
_kinit_getargs(long args, int *argc, char ***argv)
{
	PLSTATUS st;
	char *p;
	long i;

	if (args == 0) {
		*argc = 0;
		*argv = NULL;
		return PLSTATUS_SUCCESS;
	}

	/*
	 * Get size of arg strings:
	 * "arg1\0arg2\0.....argn\0\0"
	 */
	for (*argc = 0, p = (char *)args; *p != '\0'; (*argc)++) {
		while (*p != '\0') {
			p++;
		}
		p++;
	}

	pldebug("%s: Detected %d arguments\n",
	    __FUNCTION__,
	    *argc);

	/*
	 * Create the argv list.
	 */
	st = plmm_malloc(NULL, (long *)argv, (*argc) * sizeof(char *));
	if (st != PLSTATUS_SUCCESS) {
		return st;
	}

	for (i = 0, p = (char *)args; i < *argc; i++) {
		pldebug("%s: argv[%d] = \"%s\"\n",
		    __FUNCTION__,
		    i,
		    p);

		(*argv)[i] = p;
		while (*p != '\0') {
			p++;
		}
		p++;
	}

	return PLSTATUS_SUCCESS;
}


static void
_kinit_freeargs(int argc, char **argv)
{
	if (argc) {
		plmm_free(NULL, (long)argv);
	}
}



