/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/pleado/kernel/ldr.c
 * PLD loader.
 * <p>The PLD loader is thread safe but it is not signal safe</p>.
 *
 * TODO: Callbacks instead of hardcoded resolvers.
 * TODO: Lazy Bindings
 */
#include <plc.h>
#include <plspinlock.h>
#include <pllog.h>
#include <pld.h>
#include <plzma.h>

#include "kernel.h"
#include "pldnames.h"


/*
 * List of loaded modules.
 */
static PLIST mod_list = PLIST_INITIALIZED;
static PLOCK mod_list_lock = PLOCK_INITIALIZED;

#define	mod_wlock()		plock_splock(&mod_list_lock)
#define mod_rlock()		plock_splock(&mod_list_lock)
#define mod_unlock()		plock_unlock(&mod_list_lock)


/**
 * Kernel export list.
 * @see PLDEXPORT_ENTRY
 */
static PLDEXPORT_ENTRY *_kexport_list = NULL;


static void	 pldaddr2host(PLDADDR *, PLDADDR *);
static void	*pldaddr2va(PMOD *, PLDADDR *);

static PLDADDR	*export_lookup(PLDADDR *, PSEC *, PLDNAME);
static PLDADDR  *kexport_lookup(PLDADDR *, PLDNAME);

static void	 sec_release(PSEC *);
static PSEC	*sec_lookup(PMOD *, PLDNAME);
static PLSTATUS	 sec_resolve_map(PMOD *);
static PLSTATUS	 sec_resolve_reloc(PMOD *);
static PLSTATUS	 sec_resolve_export(PMOD *);
static PLSTATUS	 sec_resolve_import(PMOD *);
static PLSTATUS	 sec_resolve_nimport(PMOD *);
static PLSTATUS  sec_resolve_hash(PMOD *);
static PLSTATUS	 sec_load(PMOD *, void **);

static PLSTATUS	 mod_resolve(PMOD *);

static PLSTATUS	 _pld_open(void **, PLDFILE **);
static PLSTATUS  _pld_sym_any(void **, PMOD **, PLDNAME, int);
static PLSTATUS	 _pld_sym(void **, const void *, PLDNAME, int);


#ifdef PLDEBUG
/*
 * List of hashes.
 */
static PLIST hash_list = PLIST_INITIALIZED;
static PLOCK hash_list_lock = PLOCK_INITIALIZED;

#define	hash_wlock()		plock_splock(&hash_list_lock)
#define hash_rlock()		plock_splock(&hash_list_lock)
#define hash_unlock()		plock_unlock(&hash_list_lock)

typedef struct _PHASH {
	PLIST	 hash_list;	/* Should be the first member		*/
	PMOD	*hash_module;	/* Module containing the .hash section	*/
	PSEC	*hash_section;	/* Loaded .hash section			*/
} PHASH;

static void        hash_init();
static PLSTATUS    hash_add(const PMOD *, const PSEC *);
static void        hash_remove(const PMOD *);
static const char *hash_string(PLDNAME);


/**
 * Initialize the global hash list.
 * @see hash_list
 * @see hash_init()
 * @see hash_remove()
 * @see hash_string()
 */
static void
hash_init()
{
	PLIST_INIT(&hash_list);
	PLOCK_INIT_UNLOCKED(&hash_list_lock);
}


/**
 * Get a string representation of a specified PLDNAME.
 * @param name Peludo name whose string representation is wanted.
 * @return The string associated to the specified PLDNAME or a static
 * pointer to an empty string if that value is not present in the hash table.
 * @see PLDNAME
 * @see hash_add()
 * @see hash_init()
 * @see hash_remove()
 */
static const char *
hash_string(PLDNAME name)
{
	static char buf[1] = { 0 };
	PLDHASH_ENTRY *entry;
	PHASH *hash;
	PSEC *hsec;

	hash_rlock();

	PLIST_FOREACH(&hash_list, hash, hash_list) {

		hsec = hash->hash_section;

		for (entry = (PLDHASH_ENTRY *)hsec->se_data;
		    !PLDHASH_NULL_ENTRY(entry); entry++) {

			if (entry->hash_name == name) {
				/*
 				 * XXX
				 * Doing unlock() and returning that
				 * static pointer means we can't unload
				 * hash sections.
				 */
				hash_unlock();

				return ((const char *)
				    (long)entry->hash_string.vaddr);
			}
		}
	}

	hash_unlock();

	return buf;
}


/**
 * Add a loaded .hash section into the global hash list.
 * @param mod Module containing the .hash section;
 * @param hsec Loaded .hash section.
 * @return A peludo status code.
 * @see PLSTATUS
 * @see PMOD
 * @see PSEC
 * @see hash_init()
 * @see hash_remove()
 * @see hash_string()
 */
static PLSTATUS
hash_add(const PMOD *mod, const PSEC *hsec) 
{
	PLSTATUS st;
	PHASH *hash;

	st = plmm_malloc(NULL, (long *)&hash, sizeof(PHASH));
	if (st == PLSTATUS_SUCCESS) {
		PLIST_INIT(&hash->hash_list);
		hash->hash_module = (PMOD *)mod;
		hash->hash_section = (PSEC *)hsec;

		hash_wlock();
		plist_add_tail((PLIST *)&hash_list, (PLIST *)hash);
		hash_unlock();
	}

	return st;
}


/**
 * Remove a section from the global hash list.
 * @param mod Module containing the .hash section to remove.
 * @see PMOD
 * @see hash_list
 * @see hash_add()
 * @see hash_init()
 * @see hash_string()
 */
static void
hash_remove(const PMOD *mod)
{
	PHASH *hash;

	hash_wlock();

	PLIST_FOREACH(&hash_list, hash, hash_list) {
		if (hash->hash_module == mod) {
			plist_del((PLIST *)hash);
			(void)plmm_free(NULL, (long)hash);
			break;
		}
	}

	hash_unlock();
}
#endif	/* PLDEBUG */


/**
 * Translate a Peludo Address from network to host byte order.
 * @see PLDADDR
 */
static void
pldaddr2host(PLDADDR *dst, PLDADDR *addr)
{
	dst->paddr.addr_section = PLN32H(addr->paddr.addr_section);
	dst->paddr.addr_offset = PLN32H(addr->paddr.addr_offset);
}


/**
 * Translate a Peludo Address (host byte order) to Virtual Address.
 * @param mod Module associated to the Peludo Address;
 * @param addr Peludo Address to translate.
 * @return If the peludo address is valid into the specified module
 * its virtual address is returned; otherwise NULL.
 * @see PMOD
 * @see PLDADDR
 */
static void *
pldaddr2va(PMOD *mod, PLDADDR *addr)
{
	PSEC *sec;

	/*
	 * Get section and check offset.
	 */
	sec = sec_lookup(mod, addr->paddr.addr_section);

	if (sec == NULL || addr->paddr.addr_offset > sec->se_info.sec_size) {
#ifdef PLDEBUG
		if (sec == NULL) {
			pldebug("%s: Section not found: <%x> \"%s\"\n",
			    __FUNCTION__,
			    addr->paddr.addr_section,
			    hash_string(addr->paddr.addr_section));

		} else {
			pldebug("%s: Invalid offset: section=<%x> \"%s\", "
			    "size=%x, desired offset=%x, module=<%x> \"%s\"\n",
			    __FUNCTION__,
			    sec->se_info.sec_name,
			    hash_string(sec->se_info.sec_name),
			    sec->se_info.sec_size,
			    addr->paddr.addr_offset,
			    mod->mod_name,
			    hash_string(mod->mod_name));
		}
#endif	/* PLDEBUG */
		return NULL;
	}

	/*
	 * Translate Peludo Address to Vitual Address.
	 */
	return ((char *)sec->se_data + addr->paddr.addr_offset);
}


/**
 * Search for an exported symbol into the kernel export section.
 * @param pa Updated with the address of the exported symbol;
 * @param sym Symbol Name.
 * @return The address of the exported symbol is returned;
 * if the kernel does not have the symbol NULL is returned.
 * @see PLDADDR
 * @see PLDNAME
 * @see _kexport_list
 */
static PLDADDR *
kexport_lookup(PLDADDR *pa, PLDNAME sym)
{
	PLDEXPORT_ENTRY *entry;

	for (entry = _kexport_list; !PLDEXPORT_NULL_ENTRY(entry); entry++) {
		if (entry->exp_symname == sym) {
			/*
 			 * Don't care (yet) about .text and .data
 			 * because both sections are mapped at once.
 			 * (see comments into sec_resolve_map() and
 			 * sec_load()).
 			 */
			pa->vaddr = (uint64)(long)kbase +
			    entry->exp_symaddr.paddr.addr_offset;

			return pa;
		}
	}

	return NULL;
}


/**
 * Search for an exported symbol into a given export section.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param pa Updated with the address of the exported symbol;
 * @param exp Export setion where to search for the symbol;
 * @param sym Symbol Name.
 * @return The address of the exported symbol; if the section does not
 * have the symbol <code>NULL</code> is returned.
 * @see PLDADDR
 * @see PLDNAME
 * @see PLDSEC
 */
static PLDADDR *
export_lookup(PLDADDR *pa, PSEC *exp, PLDNAME sym)
{
	PLDEXPORT_ENTRY *entry;

	for (entry = (PLDEXPORT_ENTRY *)(exp->se_data);
	    !PLDEXPORT_NULL_ENTRY(entry); entry++) {

		if (entry->exp_symname == sym) {
			*pa = entry->exp_symaddr;
			return pa;
		}
	}

	return NULL;
}


/**
 * Release a loaded section.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param se Section to remove.
 * @see PSEC
 */
static void
sec_release(PSEC *se)
{
	long err;

	/*
	 * XXX .bss belongs to .data
	 */
	if (se->se_info.sec_name == PLDSECTION_CODE ||
	    se->se_info.sec_name == PLDSECTION_DATA) {

		if (se->se_data != NULL) {
			_plsys_pfree(&err, (long)se->se_data,
			    se->se_info.sec_size);
		}
	}

	(void)plmm_free(NULL, (long)se);
}


/**
 * Search for a section into a specified module.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module where to search;
 * @param secname Name of the desired section.
 * @return A pointer to the section or <code>NULL</code> if it does not exists.
 * @see PLDNAME
 * @see PMOD
 * @see PSEC
 */
static PSEC *
sec_lookup(PMOD *mod, PLDNAME secname)
{
	PSEC *sec;

	PLIST_FOREACH(&mod->mod_seclist, sec, se_list) {
		if (sec->se_info.sec_name == secname) {
			return sec;
		}
	}

	return NULL;
}


/**
 * Resolve a PLD hash table section.
 * @param mod Module to resolve;
 * @return A peludo status code.
 * @see PLSTATUS
 * @see PMOD
 */
static PLSTATUS
sec_resolve_hash(PMOD *mod)
{
#ifdef PLDEBUG
	PLDHASH_ENTRY *entry;
	PSEC *hsec;
	void *va;

	hsec = sec_lookup(mod, PLDSECTION_HASH);
	if (hsec == NULL) {
		return PLSTATUS_SUCCESS;
	}

	for (entry = (PLDHASH_ENTRY *)hsec->se_data;
	    !PLDHASH_NULL_ENTRY(entry); entry++) {
		/*
		 * Translate Byte order.
		 */
		entry->hash_name = PLN32H(entry->hash_name);
		pldaddr2host(&entry->hash_string, &entry->hash_string);

		/*
		 * Translate from Peludo Address to Vitual Address.
		 */
		va = pldaddr2va(mod, &entry->hash_string);
		if (va == NULL) {
			pldebug("%s: va=NULL, section=<%x> \"%s\", offset=%x\n",
			    __FUNCTION__,
			    entry->hash_string.paddr.addr_section,
			    hash_string(entry->hash_string.paddr.addr_section),
			    entry->hash_string.paddr.addr_offset);

			return PLSTATUS_INVALID_ADDRESS;
		}

/* Too noisy */
#if 0
		pldebug("%s: New hash <%x> \"%s\" at %x\n",
		    __FUNCTION__,
		    entry->hash_name,
		    hash_string(entry->hash_name),
		    (long)va);
#endif

		entry->hash_string.vaddr = (uint64)(long)va;
	}

	return hash_add(mod, hsec);
#else
	return PLSTATUS_SUCCESS;
#endif	/* PLDEBUG */
}


/**
 * Resolve memory mapping of sections.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module to resolve.
 * @return A peludo status code.
 * @see PLDSECTION_BSS
 * @see PLDSECTION_CODE
 * @see PLDSECTION_DATA
 * @see PLSTATUS
 * @see PMOD
 * @see PSEC
 */
static PLSTATUS
sec_resolve_map(PMOD *mod)
{
	long map, msiz, csiz, dsiz, bsiz, wsiz, err;
	PSEC *scode, *sdata, *sbss;
	PLSTATUS st;

	csiz = ((scode = sec_lookup(mod, PLDSECTION_CODE)) != NULL) ?
	    scode->se_info.sec_size : 0;

	dsiz = ((sdata = sec_lookup(mod, PLDSECTION_DATA)) != NULL) ?
	    sdata->se_info.sec_size : 0;

	bsiz = ((sbss = sec_lookup(mod, PLDSECTION_BSS)) != NULL) ?
	    sbss->se_info.sec_size : 0;

	if (scode == NULL && sdata == NULL) {
		return PLSTATUS_SUCCESS;
	}

	/*
	 * FIXME:
	 * Compression is supported on data sections only.
	 * This method is very memory consuming,
	 * the whole loader should be reworked.
	 */
	if (sdata->se_info.sec_flags & PLDSECTION_FL_Z) {
		long realdata;
		size_t realdatasiz;

		pldebug("%s: Calling plzma_decode for module=%s, cdata=%x, "
		    " csize=%d\n",
		    __FUNCTION__,
		    hash_string(mod->mod_name),
		    sdata->se_data,
		    dsiz);

		st = plzma_decode(&realdata, &realdatasiz, (long)sdata->se_data,
		    dsiz);

		if (st != PLSTATUS_SUCCESS) {
			return st;
		}

		sdata->se_data = (void *)realdata;
		dsiz = (long)realdatasiz;
	}

	/*
	 * .data and .bss should be 4 bytes aligned.
	 * don't ask why +4.
	 *
	 * XXX what about 64bits? See mk/script.lds
	 */
	if (bsiz != 0) {
		dsiz = PLALIGN(dsiz, 4) + 4;
	}

	wsiz = dsiz + bsiz;

	msiz = PLALIGN(csiz, PLPAGE_SIZE) + PLALIGN(wsiz, PLPAGE_SIZE);
	st = _plsys_palloc(&err, &map, msiz, PLMFLAGS_R | PLMFLAGS_W);
	if (st != PLSTATUS_SUCCESS) {
		return st;
	}

	/*
	 * Copy code.
	 */
	if (csiz != 0) {
		plsys_memmove((void *)map, scode->se_data, csiz);
		scode->se_data = (void *)map;
		scode->se_info.sec_size = PLALIGN(csiz, PLPAGE_SIZE);
	}

	/*
	 * Copy data.
	 */
	if (dsiz != 0) {
		plsys_memmove((void *)(map + PLALIGN(csiz, PLPAGE_SIZE)),
		    sdata->se_data, dsiz);

		if (sdata->se_info.sec_flags & PLDSECTION_FL_Z) {
			plmm_free(&err, (long)sdata->se_data);
		}

		sdata->se_data = (void *)(map + PLALIGN(csiz, PLPAGE_SIZE));
	}

	/*
	 * Zeroe bss.
	 */
	if (bsiz != 0) {
		plsys_memset((void *)(map + PLALIGN(csiz, PLPAGE_SIZE) + dsiz)
		    , 0, bsiz);

		sbss->se_data = (void *)(map + PLALIGN(csiz, PLPAGE_SIZE) +
		    dsiz);
	}

	/*
	 * Expand .data to include .bss.
	 */
	if (wsiz != 0) {
		sdata->se_info.sec_size = PLALIGN(wsiz, PLPAGE_SIZE);
	}

	pldebug("%s: <%x> \"%s\": .text=%x/%x .data=%x/%x .bss=%x/%x\n",
	    __FUNCTION__,
	    mod->mod_name,
	    hash_string(mod->mod_name),
	    (scode != NULL ? scode->se_data : 0),
	    csiz,
	    (sdata != NULL ? sdata->se_data : 0),
	    dsiz,
	    (sdata != NULL ? ((uint8 *)sdata->se_data) + dsiz: 0),
	    bsiz);

	return PLSTATUS_SUCCESS;
}


/**
 * Resolve Relocations.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module to resolve.
 * @return A peludo status code.
 * @see PLDRELOC_ENTRY
 * @see PLDSECTION_RELOC
 * @see PMOD
 * @see PSEC
 */
static PLSTATUS
sec_resolve_reloc(PMOD *mod)
{
	PLDRELOC_ENTRY *entry;
	void *symva, *torelva;
	PLDADDR sym, torel;
	PSEC *reloc;

	reloc = sec_lookup(mod, PLDSECTION_RELOC);
	if (reloc == NULL) {
		return PLSTATUS_SUCCESS;
	}

	for (entry = (PLDRELOC_ENTRY *)reloc->se_data;
	    !PLDRELOC_NULL_ENTRY(entry); entry++) {
		/*
		 * Get Addresses.
		 */
		pldaddr2host(&sym, &entry->rel_symbol);
		pldaddr2host(&torel, &entry->rel_address);
		entry->rel_type = PLN32H(entry->rel_type);

		symva = pldaddr2va(mod, &sym);
		torelva = pldaddr2va(mod, &torel);

		if (symva == NULL || torelva == NULL) {

			pldebug("%s: Failed: module=<%x> \"%s\"\n",
			    "\tsymva=%x, section=<%x> \"%s\", offset=%x\n"
			    "\ttorelva=%x section=<%s> \"%s\" offset=%x\n",
			    __FUNCTION__,
			    mod->mod_name,
			    hash_string(mod->mod_name),

			    symva,
			    sym.paddr.addr_section,
			    hash_string(sym.paddr.addr_section),
			    sym.paddr.addr_offset,

			    torelva,
			    torel.paddr.addr_section,
			    hash_string(torel.paddr.addr_section),
			    torel.paddr.addr_offset);

			return PLSTATUS_INVALID_ADDRESS;
		}

		/*
		 * Apply relocation.
		 */
		switch (entry->rel_type) {
		case PLDRELOC_TYPE_ABS:
			*(void **)torelva = symva;
			break;

		case PLDRELOC_TYPE_ADD:
			*(void **)torelva += (long)symva;
			break;

		default:;
		}
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Resolve Exports.
 * <p>Translate Peludo Address of export table to Virtual Memory Address.</p>
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module to resolve.
 * @return A peludo status code.
 * @see PLDEXPORT_ENTRY
 * @see PLDSECTION_EXPORT
 * @see PLSTATUS
 * @see PMOD
 * @see PSEC
 */
static PLSTATUS
sec_resolve_export(PMOD *mod)
{
	PLDEXPORT_ENTRY *entry;
	PSEC *exp;
	void *va;

	exp = sec_lookup(mod, PLDSECTION_EXPORT);
	if (exp == NULL) {
		return PLSTATUS_SUCCESS;
	}

	for (entry = (PLDEXPORT_ENTRY *)exp->se_data;
	    !PLDEXPORT_NULL_ENTRY(entry); entry++) {
		/*
		 * Translate Byte order.
		 */
		entry->exp_symname = PLN32H(entry->exp_symname);
		pldaddr2host(&entry->exp_symaddr, &entry->exp_symaddr);

		/*
		 * Translate from PLD Address to Vitual Address.
		 */
		va = pldaddr2va(mod, &entry->exp_symaddr);
		if (va == NULL) {

			pldebug("%s: va=NULL, symbol=<%x> \"%s\", "
			    "section=<%x> \"%s\", offset=%x\n",
			    __FUNCTION__,
			    entry->exp_symname,
			    hash_string(entry->exp_symname),
			    entry->exp_symaddr.paddr.addr_section,
			    hash_string(entry->exp_symaddr.paddr.addr_section),
			    entry->exp_symaddr.paddr.addr_offset);

			return PLSTATUS_INVALID_ADDRESS;
		}

		pldebug("%s: Exporting <%x> \"%s\"\n",
		    __FUNCTION__,
		    entry->exp_symname,
		    hash_string(entry->exp_symname));

		entry->exp_symaddr.vaddr = (uint64)(long)va;
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Resolve Imports.
 * <p>The current implementation supports only one Imported Module entry
 * (the wildcard) to emulate the ELF way to get imports.</p>
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module to resolve.
 * @return A peludo status code.
 * @see PLDIMPORT_ENTRY
 * @see PLDIMPORT_SYMBOL
 * @see PLSTATUS
 * @see PMOD
 * @see PSEC
 */
static PLSTATUS
sec_resolve_import(PMOD *mod)
{
	PLDIMPORT_ENTRY *ientry;
	PLDIMPORT_SYMBOL *isym;
	void *addr, *iaddr;
	PLSTATUS st;
	PSEC *sec;

	/*
	 * Get Import Section.
	 */
	sec = sec_lookup(mod, PLDSECTION_IMPORT);
	if (sec == NULL) {
		/*
		 * No imports.
		 */
		return PLSTATUS_SUCCESS;
	}

	ientry = (PLDIMPORT_ENTRY *)sec->se_data;

	/*
	 * Transalte byte order.
	 */
	ientry->imp_modname = PLN32H(ientry->imp_modname);
	pldaddr2host(&ientry->imp_symbols, &ientry->imp_symbols);

	if (PLDADDR_NULL(&ientry->imp_symbols)) {
		/*
		 * No Import Symbol Table.
		 */
		return PLSTATUS_SUCCESS;
	}

	/*
	 * Support only module -1 (the wildcard).
	 */
	if (ientry->imp_modname != 0xFFFFFFFF) {
		/* FIXME */
		pldebug("%s: Non wildcard modules not supported yet\n",
		    __FUNCTION__);

		return PLSTATUS_INVALID;
	}

	/*
	 * Process the Import Symbol Table.
	 */
	isym = (PLDIMPORT_SYMBOL *)pldaddr2va(mod, &ientry->imp_symbols);
	if (isym == NULL) {
		return PLSTATUS_INVALID_ADDRESS;
	}

	while (!PLDIMPORT_NULL_SYMBOL(isym)) {
		/*
		 * Translate byte order and the get virtual address of the
		 * address to update with the address of the imported symbol.
		 */
		isym->imp_symname = PLN32H(isym->imp_symname);
		pldaddr2host(&isym->imp_symaddr, &isym->imp_symaddr);

		pldebug("%s: Importing <%x> \"%s\"\n",
		    __FUNCTION__,
		    isym->imp_symname,
		    hash_string(isym->imp_symname));

		iaddr = pldaddr2va(mod, &isym->imp_symaddr);
		if (iaddr == NULL) {
			/*
			 * Invalid address: Malformed table.
			 */
			pldebug("%s: NULL symbol address: %x/%x, "
			    "<%s> \"%s\": <%d> \"%s\".\n",
			    __FUNCTION__,
			    isym->imp_symaddr.paddr.addr_section,
			    isym->imp_symaddr.paddr.addr_offset,
			    isym->imp_symname,
			    hash_string(isym->imp_symname));

			return PLSTATUS_MALFORMED_TABLE;
		}

		/*
		 * Search for the symbol on ALL modules
		 * and get its virtual address.
		 */
		st = _pld_sym_any(&addr, NULL, isym->imp_symname, 0);
		if (st != PLSTATUS_SUCCESS) {
			/*
			 * Symbol not found: Dependency broken.
			 */
			return st;
		}

		/*
		 * Update the importing entry with the
		 * virtual address of the imported symbol.
		 */
		*(void **)iaddr = addr;

		isym++;
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Load symbols from a native DLL.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module containing the Native Import entry;
 * @param dll Base address of the loaded DLL;
 * @param isym Null terminated array of imported symbols (PLD formatted).
 * @return A peludo status code.
 * @see PMOD
 * @see PLDNIMPORT_SYMBOL
 * @see PLSTATUS
 * @see pld_sym_dlsym()
 */
static PLSTATUS
sec_nimport_dlsym(PMOD *mod, long dll, PLDNIMPORT_SYMBOL *isym)
{
	long iaddr, addr;
	char *symname;
	PLSTATUS st;

	while (1) {
		pldaddr2host(&isym->nimp_symname, &isym->nimp_symname);
		pldaddr2host(&isym->nimp_symaddr, &isym->nimp_symaddr);

		if (PLDADDR_NULL(&isym->nimp_symname) &&
		    PLDADDR_NULL(&isym->nimp_symaddr)) {
			/*
			 * End of import table.
			 */
			break;
		}

		symname = (char *)pldaddr2va(mod, &isym->nimp_symname);
		iaddr = (long)pldaddr2va(mod, &isym->nimp_symaddr);
		if (symname == NULL || iaddr == 0) {

			pldebug("%s: symname = %x, iaddr = %x\n",
			    __FUNCTION__,
			    symname, iaddr);

			/*
			 * Invalid address: Malformed table.
			 */
			return PLSTATUS_MALFORMED_TABLE;
		}

		/*
		 * Get the virtual address of the imported symbol.
		 */
		st = plsys_dlsym(&addr, dll, symname);
		if (st != PLSTATUS_SUCCESS) {
			/*
			 * Symbol not found. Dependency broken.	
			 */
			return st;
		} 

		/*
		 * Update the importing entry with the
		 * virtual address of the imported symbol.
		 */
		*(long **)iaddr = (long *)addr;

		isym++;
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Resolve Native Imports.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module to resolve.
 * @return A peludo status code.
 * @see PLDNIMPORT_ENTRY
 * @see PLSTATUS
 * @see PMOD
 * @see PSEC
 * @see plsys_dlopen()
 * @see sec_nimport_dlsym()
 */
static PLSTATUS
sec_resolve_nimport(PMOD *mod)
{
	PLDNIMPORT_ENTRY *ientry;
	PLDNIMPORT_SYMBOL *isym;
	PLSTATUS st;
	PSEC *sec;
	char *dllname;
	long dll;

	/*
	 * Get Native Import Section.
	 */
	sec = sec_lookup(mod, PLDSECTION_NIMPORT);
	if (sec == NULL) {
		/*
		 * No Native Imports.
		 */
		return PLSTATUS_SUCCESS;
	}

	ientry = (PLDNIMPORT_ENTRY *)sec->se_data;
	while (1) {
		pldaddr2host(&ientry->nimp_dllname, &ientry->nimp_dllname);
		pldaddr2host(&ientry->nimp_symbols, &ientry->nimp_symbols);

		if (PLDADDR_NULL(&ientry->nimp_dllname)) {
			/*
			 * End of DLL table.
			 */
			break;
		}

		/*
		 * Load DLL.
		 */
		if (PLDADDR_WILDCARD(&ientry->nimp_dllname)) {
			/*
			 * Wildcard DLL.
			 */
			dll = 0;

		} else {
			dllname = pldaddr2va(mod, &ientry->nimp_dllname);
			if (dllname == NULL) {
				/*
				 * Invalid DLL name.
				 */
				return PLSTATUS_MALFORMED_TABLE;
			}

			st = plsys_dlopen(&dll, dllname);
			if (st != PLSTATUS_SUCCESS) {
				/*
				 * DLL not found. Dependency broken.
				 */
				return st;
			}
		}

		/*
		 * Import symbols.
		 */
		if (!PLDADDR_NULL(&ientry->nimp_symbols)) {
			isym = pldaddr2va(mod, &ientry->nimp_symbols);
			if (isym == NULL) {
				/*
				 * Invalid address: Malformed table.
				 */
				return PLSTATUS_MALFORMED_TABLE;
			}

			st = sec_nimport_dlsym(mod, dll, isym);
			if (st != PLSTATUS_SUCCESS) {
				/*
				 * Symbol not found. Dependency broken.
				 */
				return st;
			}
		}

		ientry++;
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Set correct page permissions for each section.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod (IN) Module to resolve.
 * @return A peludo status code.
 * @see PLDSECTION_FLAGS
 * @see PLSTATUS
 * @see PMOD
 */
static PLSTATUS
sec_resolve_perms(PMOD *mod)
{
	PLMFLAGS mflags;
	PLSTATUS st;
	PSEC *sec;
	long err;

	st = PLSTATUS_SUCCESS;

	PLIST_FOREACH(&mod->mod_seclist, sec, se_list) {
		if (sec->se_info.sec_name == PLDSECTION_CODE ||
		    sec->se_info.sec_name == PLDSECTION_DATA) {

			mflags = 0;

			if (sec->se_info.sec_flags & PLDSECTION_FL_X) {
				mflags |= PLMFLAGS_X;
			}

			if (sec->se_info.sec_flags & PLDSECTION_FL_R) {
				mflags |= PLMFLAGS_R;
			}

			if (sec->se_info.sec_flags & PLDSECTION_FL_W) {
				mflags |= PLMFLAGS_W;
			}

			st = _plsys_mprotect(&err, (long)sec->se_data,
			    sec->se_info.sec_size, mflags);

			if (st != PLSTATUS_SUCCESS) {
				break;
			}
		}
	}

	return st;
}


/**
 * Load a section.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Pointer to the module that owns the section.
 * @param sec Section to load; at exit it is updated with the pointer to the
 * next section (or whatever continues).
 * @return A peludo status code.
 * @see PLDSECTION
 * @see PLSTATUS
 * @see PMOD
 * FIXME: This function is a hack to support only one mmap for .text and .data.
 *      The loader should be section independent and use a callback table 
 *	to load/resolve each type of section.
 */
static PLSTATUS
sec_load(PMOD *mod, void **sec)
{
	PLDSECTION *fsec;
	PLSTATUS st;
	PSEC *msec;
	int msize;

	fsec = *(PLDSECTION **)sec;

	msize = sizeof(PSEC);

	pldebug("%s: Loading <%x> \"%s\"\n",
	    __FUNCTION__,
	    PLN32H(fsec->sec_name),
	    hash_string(PLN32H(fsec->sec_name)));

	switch (PLN32H(fsec->sec_name)) {
	case PLDSECTION_CODE:
	case PLDSECTION_DATA:
	case PLDSECTION_BSS:
		break;

	case PLDSECTION_EXPORT:
	case PLDSECTION_IMPORT:
	case PLDSECTION_NIMPORT:
	case PLDSECTION_RELOC:
	case PLDSECTION_HASH:
		msize += PLN32H(fsec->sec_size);
		break;

	default:
		return PLSTATUS_NOT_IMPLEMENTED;
	}

	st = plmm_malloc(NULL, (long *)&msec, msize);
	if (st != PLSTATUS_SUCCESS) {
		return st;
	}

	PLIST_INIT(&msec->se_list);

	// msec->se_info = *fsec;
	msec->se_info.sec_name = PLN32H(fsec->sec_name);
	msec->se_info.sec_size = PLN32H(fsec->sec_size);
	msec->se_info.sec_flags = PLN32H(fsec->sec_flags);

	if (msec->se_info.sec_name == PLDSECTION_CODE ||
	    msec->se_info.sec_name == PLDSECTION_DATA ||
	    msec->se_info.sec_name == PLDSECTION_BSS) {
		/*
		 * The resolver should mmap:
		 * This is a hack to support mapping of both sections at once.
		 */
		msec->se_data = (void *)(fsec + 1);

	} else {
		/*
		 * In the other sections the contents
		 * continue after the PSEC header.
		 */
		msec->se_data = (void *)(msec + 1);

		/*
		 * Once copied appropiate resolvers should translate byteorder.
		 */
		plsys_memmove(msec->se_data, (void *)(fsec + 1),
		    msec->se_info.sec_size);
	}

	/*
	 * Add section into the list.
	 */
	plist_add_tail((PLIST *)&mod->mod_seclist, (PLIST *)msec);

	/*
	 * Point to the next section.
	 */
	*(char **)sec += sizeof(PLDSECTION);

	/*
	 * A .bss Section only give the .bss memory size
	 * but there is no data in the PLDSECTION
	 * so we directly go to the next Section.
	 * FIXME; this violates the PLD spec. elf2pld should also be fixed.
	 */
	if (msec->se_info.sec_name != PLDSECTION_BSS) {
		*(char **)sec += msec->se_info.sec_size;
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Remove a module from the list and release used resources.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module to remove;
 * @param inlist Boolean value indicating whether the module was
 * added into the list of modules.
 * @see PMOD
 * @see mod_list
 */
void
mod_release(PMOD *mod, int inlist)
{
	PSEC *sec;

	/*
	 * Remove module from the list of modules.
	 */
	if (inlist) {
		plist_del(&mod->mod_list);
	}

#ifdef PLDEBUG
	/*
	 * Release the hash entry associated to the module.
	 */
	hash_remove(mod);
#endif

	/*
	 * Release all sections.
	 */
	while (!plist_empty(&mod->mod_seclist)) {
		sec = (PSEC *)mod->mod_seclist.next;
		plist_del((PLIST *)sec);
		sec_release(sec);
	}

	(void)plmm_free(NULL, (long)mod);
}


/**
 * Search for a loaded module.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param name Name of the module (can't be NULL).
 * @return A pointer to the loaded module or <code>NULL</code> if it is not
 * loaded.
 * @see PLDNAME
 * @see PMOD
 * @see mod_list
 * @see mod_list_lock
 */
static PMOD *
mod_lookup(PLDNAME name)
{
	PMOD *mod;

	PLIST_FOREACH(&mod_list, mod, mod_list) {
		if (mod->mod_name == name) {
			return mod;
		}
	}

	return NULL;
}


/**
 * Resolve a module.
 * <p>Resolve sections of a module.</p>
 * <p>The list of modules should be locked before calling this function.</p>
 * @param mod Module to resolve.
 * @return A peludo status code.
 * @see PLSTATUS
 * @see PMOD
 * @see sec_resolve_export()
 * @see sec_resolve_import()
 * @see sec_resolve_hash()
 * @see sec_resolve_map()
 * @see sec_resolve_nimport()
 * @see sec_resolve_perms()
 * @see sec_resolve_reloc()
 */
static PLSTATUS
mod_resolve(PMOD *mod)
{
	int st;

	if ((st = sec_resolve_hash(mod)) == PLSTATUS_SUCCESS &&
	    (st = sec_resolve_map(mod)) == PLSTATUS_SUCCESS &&
	    (st = sec_resolve_reloc(mod)) == PLSTATUS_SUCCESS &&
	    (st = sec_resolve_export(mod)) == PLSTATUS_SUCCESS &&
	    (st = sec_resolve_import(mod)) == PLSTATUS_SUCCESS &&
	    (st = sec_resolve_nimport(mod)) == PLSTATUS_SUCCESS &&
	    (st = sec_resolve_perms(mod)) == PLSTATUS_SUCCESS);

	return st;
}


/**
 * Get the virtual address of a symbol.
 * <p>Traverse all modules searching for the desired symbol.</p>
 * @param addr Pointer to be updated with the virtual address;
 * if this parameter is <code>NULL</code> this function verifies the
 * symbol existence;
 * @param mod If not <code>NULL</code> it is updated with the module
 * that exports the symbol;
 * @param sym Symbol Name;
 * @param dolock If non-zero the module list should be locked
 * (if zero the module list is already locked by the caller).
 * @return A peludo status code.
 * @see PLDNAME
 * @see PLSTATUS
 * @see _pld_sym()
 */
static PLSTATUS
_pld_sym_any(void **addr, PMOD **emod, PLDNAME sym, int dolock)
{
	PLSTATUS st;
	PMOD *mod;

	/*
	 * Look at modules exports.
	 */
	st = PLSTATUS_SYMBOL_NOT_FOUND;

	if (dolock) {
		mod_rlock();
	}

	PLIST_FOREACH(&mod_list, mod, mod_list) {
		if ((st = _pld_sym(addr, mod, sym, 0)) == PLSTATUS_SUCCESS) {
			break;
		}
	}

	if (dolock) {
		mod_unlock();
	}

	/*
	 * Look at the kernel exports.
	 */
	if (st == PLSTATUS_SYMBOL_NOT_FOUND) {
		st = _pld_sym(addr, NULL, sym, dolock);
		mod = NULL;
	}

#ifdef PLDEBUG
	if (st == PLSTATUS_SYMBOL_NOT_FOUND) {
		pldebug("%s: Symbol Not Found <%x> \"%s\"\n",
		    __FUNCTION__,
		    sym,
		    hash_string(sym));
	} else if (st == PLSTATUS_SUCCESS) {
		pldebug("%s: Symbol <%x> \"%s\" found into module "
		    "<%x> \"%s\"\n",
		    __FUNCTION__,
		    sym,
		    hash_string(sym),
		    ((mod == NULL) ? 0 : mod->mod_name),
		    ((mod == NULL) ? "KERNEL" : hash_string(mod->mod_name)));
	}
#endif

	if (st == PLSTATUS_SUCCESS && emod != NULL) {
		*emod = mod;
	}

	return st;
}


/**
 * Install (locate, resolve and initialize) a PLD Module.
 * <p>If the module exports the <code>_init(void)</code> function
 * it is called.</p>
 * @param h If not <code>NULL</code> it is updated with the handle to
 * the installed module;
 * @param file PLD File to install; at exit it is updated
 * with a pointer to the next PLDFILE (or whatever follows).
 * @return A peludo status code.
 * @see PLDFILE
 * @see PLSTATUS
 * @see mod_lookup()
 * @see mod_resolve()
 * @see pld_sym()
 * @see sec_load()
 */
static PLSTATUS
_pld_open(void **h, PLDFILE **file)
{
	void (*_init)();
	PLDFILE *pld;
	PLSTATUS st;
	void *next;
	PMOD *mod;
	uint32 i;

	if (file == NULL || *file == NULL) {
		return PLSTATUS_SUCCESS;
	}

	st = PLSTATUS_SUCCESS;
	pld = *file;

	mod_wlock();

	pldebug("%s: Opening <%x> \"%s\"\n",
	    __FUNCTION__,
	    PLN32H(pld->pld_name),
	    hash_string(PLN32H(pld->pld_name)));

	if ((mod = mod_lookup(PLN32H(pld->pld_name))) != NULL) {
		/*
		 * There is another module using the same name.
		 */
		st = PLSTATUS_MODULE_NAME_DUPLICATED;

	} else if ((st = plmm_malloc(NULL, (long *)&mod, sizeof(PMOD)))
	    != PLSTATUS_SUCCESS) {
		goto bye;

	} else {
		/*
		 * Initialize module members.
		 */
		mod->mod_name = PLN32H(pld->pld_name);
		PLIST_INIT(&mod->mod_seclist);

		/*
		 * Point to the next section (or whatever is there).
		 */
		next = (void *)(pld + 1);

		/*
		 * Load sections into memory.
		 */
		for (i = 0; i < PLN32H(pld->pld_sections); i++) {
			if ((st = sec_load(mod, &next)) != PLSTATUS_SUCCESS) {
fail:				mod_release(mod, 0);
				goto bye;
			}
		}

		/*
		 * Resolve sections.
		 */
		if ((st = mod_resolve(mod)) != PLSTATUS_SUCCESS) {
			goto fail;
		}

		/*
		 * Point to the next PLDFILE (or whatever is there).
		 */
		*file = next;

		/*
		 * Add module to the list.
		 */
		plist_add_tail((PLIST *)&mod_list, (PLIST *)mod);
	}

bye:	mod_unlock();

	if (st == PLSTATUS_SUCCESS) {
		/*
		 * Call _init()
		 */
		if (pld_sym((void **)&_init, mod, PLDNAME_SYMBOL__INIT)
		    == PLSTATUS_SUCCESS) {

			pldebug("%s: Calling _init() from <%x> \"%s\"\n",
			    __FUNCTION__,
			    mod->mod_name,
			    hash_string(mod->mod_name));

			_init();

			pldebug("%s: _init() returned\n", __FUNCTION__);

		} else {
			pldebug("%s: _init() not found into <%x> \"%s\"\n",
			    __FUNCTION__,
			    mod->mod_name,
			    hash_string(mod->mod_name));
		}

		/*
		 * Set Module Handle.
		 */
		if (h != NULL) {
			*h = (void *)mod;
		}
	}

	return st;
}


/**
 * Get virtual address of a symbol.
 * <p>The list of modules should be locked before calling this function.</p>
 * @param addr Updated with the virtual address of the desired symbol;
 * if this parameter is <code>NULL</code> this function just verifies the
 * symbol existence;
 * @param h Handle to the PLD module exporting the desired symbol
 * (use <code>NULL</code> to get symbols exported by the Peludo Kernel);
 * @param sym Symbol Name;
 * @param dolock If non zero the module list should be locked
 * (if zero means the module list is already locked by the caller).
 * @return A peludo status code.
 * @see PLDNAME
 * @see sec_lookup()
 * @see export_lookup()
 * @see kexport_lookup()
 */
static PLSTATUS
_pld_sym(void **addr, const void *h, PLDNAME sym, int dolock)
{
	PLDADDR pa;
	PSEC *exp;
	PLSTATUS st;

	st = PLSTATUS_SUCCESS;

	if (h == NULL) {
		/*
		 * NULL is the kernel handle.
		 */
		if (kexport_lookup(&pa, sym) == NULL) {
			st = PLSTATUS_SYMBOL_NOT_FOUND;
		} else {
			pldebug("%s: Found <%x> \"%s\" at %x into kernel\n",
			    __FUNCTION__,
			    sym,
			    hash_string(sym),
			    (long)pa.vaddr);
		}

	} else {
		PMOD *mod = (PMOD *)h;

		if (dolock) {
			mod_rlock();
		}

		exp = sec_lookup(mod, PLDSECTION_EXPORT);

		if (exp == NULL || export_lookup(&pa, exp, sym) == NULL) {
			st = PLSTATUS_SYMBOL_NOT_FOUND;
		} else {
			pldebug("%s: Found <%x> \"%s\" at %x into <%x> "
			    "\"%s\"\n",
			    __FUNCTION__,
			    sym,
			    hash_string(sym),
			    (long)pa.vaddr,
			    mod->mod_name,
			    hash_string(mod->mod_name));
		}

		if (dolock) {
			mod_unlock();
		}
	}

	if (st == PLSTATUS_SUCCESS && addr != NULL) {
		*addr = (void *)(long)(pa.vaddr);
	}

	return st;
}


/**
 * Initialize the PLD loader.
 * @param esec Address of the kernel export section (in PLD format);
 * updated to the address that follows the section.
 * @return A peludo status code.
 * @see PLSTATUS
 * @see pld_release()
 */
PLSTATUS PLEXPORT
pld_init(long *esec)
{
	PLDEXPORT_ENTRY *e;
	PLDSECTION *sec;
	PLSTATUS st;
	size_t size;
#ifdef PLDEBUG
	PLDHASH_ENTRY *he;
	PSEC *psec;
#endif

	pldebug("%s: Initializing PLD Loader\n", __FUNCTION__);

	PLIST_INIT(&mod_list);
	PLOCK_INIT_UNLOCKED(&mod_list_lock);

#ifdef PLDEBUG
	hash_init();
#endif

	if (esec != NULL && *esec != 0) {
		/*
		 * Get Kernel Export Table.
		 */
		pldebug("%s: Initializing Peludo Export Table:\n",
		    __FUNCTION__);

		sec = (PLDSECTION *)*esec;
		size = PLN32H(sec->sec_size);

		st = plmm_malloc(NULL, (long *)&_kexport_list, size);
		if (st != PLSTATUS_SUCCESS) {
			return st;
		}

		plsys_memmove((void *)_kexport_list, (void *)(sec + 1), size);
		*esec += sizeof(PLDSECTION) + size;

#ifdef PLDEBUG
		/*
		 * Get the Kernel Hash Table (located after the export table).
		 */
		sec = (PLDSECTION *)*esec;
		size = PLN32H(sec->sec_size);
		st = plmm_malloc(NULL, (long *)&psec, sizeof(PSEC) + size);
		if (st != PLSTATUS_SUCCESS) {
			return st;
		}

		PLIST_INIT(&psec->se_list);
		psec->se_info.sec_name = PLN32H(sec->sec_name);
		psec->se_info.sec_size = size;
		psec->se_info.sec_flags = PLN32H(sec->sec_flags);
		psec->se_data = (void *)(psec + 1);

		plsys_memmove(psec->se_data, (void *)(sec + 1), size);
		*esec += sizeof(PLDSECTION) + size;

		for (he = (PLDHASH_ENTRY *)psec->se_data;
		    !PLDHASH_NULL_ENTRY(he); he++) {
			/*
			 * Translate Byte order.
			 */
			he->hash_name = PLN32H(he->hash_name);
			pldaddr2host(&he->hash_string, &he->hash_string);

			/*
			 * Translate from Peludo Address to Vitual Address.
			 */
			he->hash_string.vaddr = (uint64)(long)
			    psec->se_data + he->hash_string.paddr.addr_offset;
		}

		hash_add(NULL, psec);
#endif	/* PLDEBUG */

		/*
		 * Translate Exports from PLD Address to Vitual Address.
		 */
		for (e = _kexport_list; !PLDEXPORT_NULL_ENTRY(e); e++) {
			e->exp_symname = PLN32H(e->exp_symname);
			pldaddr2host(&e->exp_symaddr, &e->exp_symaddr);

			pldebug("\tKernel Export <%x> \"%s\" at %x\n",
			    e->exp_symname,
			    hash_string(e->exp_symname),
			    kbase + e->exp_symaddr.paddr.addr_offset);
		}
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Uninstall all PLD modules.
 * <p>Uninstall all modules (call also _fini() function of modules exporting
 * it). The prototype is: <code>void _fini(void)</code>.</p>
 *
 * <p>Since <code>pld_release()</code> calls <code>_fini()</code> when the
 * module list is locked that <code>_fini()</code> function should not call
 * any of the <code>pld_xxx()</code> functions otherwise it deadlocks.
 * </p>
 *
 * @see mod_release()
 * @see pld_init()
 */
void PLEXPORT
pld_release()
{
	void (*_fini)();
	PLDADDR addr;
	PSEC *exp;

	pldebug("%s: Releasing PLD Loader\n", __FUNCTION__);

	mod_wlock();

	while (!plist_empty(&mod_list)) {
		/*
		 * Call _fini()
		 */
		exp = sec_lookup((PMOD *)mod_list.prev, PLDSECTION_EXPORT);
		if (exp != NULL &&
		    export_lookup(&addr, exp, PLDNAME_SYMBOL__FINI) != NULL) {
			_fini = (void *)(long)addr.vaddr;
			_fini();
		}

		/*
		 * Release module.
		 */
		mod_release((PMOD *)mod_list.prev, 1);
	}

	if (_kexport_list != NULL) {
		(void)plmm_free(NULL, (long)_kexport_list);
	}

#ifdef PLDEBUG
	(void)plmm_free(NULL, (long)(((PHASH *)hash_list.next)->hash_section));
	hash_remove(NULL);
#endif
}


/**
 * Install (locate, resolve and initialize) a PLD chain.
 * <p>The PLD Chain must end with a 0 filled PLDFILE entry.</p>
 * @param file Chain of PLDs to install.
 * @return A peludo status code.
 * @see PLDFILE
 * @see PLSTATUS
 * @see _pld_open()
 */
PLSTATUS PLEXPORT
pld_open_all(const PLDFILE *file)
{
	PLDFILE *pld;
	PLSTATUS st;

	st = PLSTATUS_SUCCESS;

	if (file != NULL) {

		pld = (PLDFILE *)file;

		while (!PLDFILE_NULL_ENTRY(pld)) {
			if ((st = _pld_open(NULL, &pld)) != PLSTATUS_SUCCESS) {
				break;
			}
		}
	}

	return st;
}


/**
 * Install (locate, resolve and initialize) a PLD Module.
 * @param h If not <code>NULL</code> it is updated with the handle
 * to the installed module;
 * @param file PLD File to install.
 * @return A peludo status code.
 * @see PLDFILE
 * @see PLSTATUS
 * @see _pld_open()
 */
PLSTATUS PLEXPORT
pld_open(void **h, const PLDFILE *file)
{
	PLDFILE *pld = (PLDFILE *)file;
	return ((file == NULL) ? PLSTATUS_SUCCESS : _pld_open(h, &pld));
}


/**
 * Get the virtual address of a symbol.
 * <p>Traverse all modules searching for the desired symbol.</p>
 * @param addr Updated with the virtual address; if this parameter is
 * <code>NULL</code> this function verifies the symbol existence;
 * @param h If not <code>NULL</code> it is updated with the handler
 * of the module that exports the symbol;
 * @param sym Symbol Name.
 * @return A peludo status code.
 * @see PLDNAME
 * @see PLSTATUS
 * @see _pld_sym_any()
 */
PLSTATUS PLEXPORT
pld_sym_any(void **addr, void **h, PLDNAME sym)
{
	return _pld_sym_any(addr, (PMOD **)h, sym, 1);
}


/**
 * Get the virtual address of a symbol.
 * @param addr Updated with the virtual address; if this parameter is
 * <code>NULL</code> this function verifies the symbol existence;
 * @param h Handle to the PLD module exporting the desired symbol
 * (use <code>NULL</code> to get symbols exported by the Peludo Kernel);
 * @param sym Symbol Name.
 * @return A peludo status code.
 * @see PLDNAME
 * @see PLSTATUS
 * @see _pld_sym()
 */
PLSTATUS PLEXPORT
pld_sym(void **addr, const void *h, PLDNAME sym)
{
#ifdef PLDEBUG
	PLSTATUS st;

	st = _pld_sym(addr, h, sym, 1);
	if (st == PLSTATUS_SYMBOL_NOT_FOUND) {
		pldebug("%s: Symbol Not Found <%x> \"%s\"\n",
		    __FUNCTION__,
		    sym,
		    hash_string(sym));
	}

	return st;

#else
	return _pld_sym(addr, h, sym, 1);

#endif
}


/**
 * Get a section of a loaded module.
 * @param sec Updated with the address of the desired section;
 * if this parameter is <code>NULL</code> this function verifies
 * the section existence;
 * @param h Handle to the PLD module;
 * @param secname Section name.
 * @return A peludo status code.
 * @see PLDNAME
 * @see PLSTATUS
 * @see PSEC
 */
PLSTATUS PLEXPORT
pld_getsec(PSEC **sec, const void *h, PLDNAME secname)
{
	PSEC *s;

	if (h != NULL) {
		mod_rlock();
		s = sec_lookup((PMOD *)h, secname);
		mod_unlock();

		if (sec != NULL) {
			*sec = s;
		}

		return ((s != NULL) ?
		    PLSTATUS_SUCCESS : PLSTATUS_SECTION_NOT_FOUND);
	}

	/* NULL module is invalid */
	return PLSTATUS_INVALID;
}


/**
 * Get the address of a PLD's .data section.
 * @param h Handle to the PLD module;
 * @param size updated with the size of the .data section.
 * @return On success the virtual address of the data section;
 * on error <code>NULL</code>.
 */
void PLEXPORT *
pld_getdata(const void *h, size_t *size)
{
	PSEC *sec;

	if (pld_getsec(&sec, h, PLDSECTION_DATA) == PLSTATUS_SUCCESS) {
		if (size != NULL) {
			*size = sec->se_info.sec_size;
		}

		return sec->se_data;
	}

	return NULL;
}


/**
 * Get the handle of a loaded PLD module.
 * @param name Name of loaded PLD module.
 * @return If the module exists a handle to it is returned, otherwise
 * <code>NULL</code> is returned.
 * @see PLDNAME
 */
void PLEXPORT *
pld_lookup(PLDNAME name)
{
	void *modh;

	if (name != 0) {
		mod_rlock();
		modh = mod_lookup(name);
		mod_unlock();
	} else {
		modh = NULL;
	}

	return modh;
}


