/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/include/plc.h
 * FIXME: Remove this header. Move plsys/plxxx to plsys.h
 */
#ifndef PLC_H
#define PLC_H

#include <plsysdep.h>
#include <pldef.h>
#include <plist.h>


/*
 * Stringfication.
 */
#define __plstr(x)	#x


/*
 * Crappy STDARG.
 */
#if defined(PLARCH_i386)

typedef uint32 * plva_list;

#define plva_arg(va, type)					\
	((type)*((va += (sizeof(type) >> 2)) - (sizeof(type) >> 2)))

#define plva_start(va, last)					\
do {								\
	__asm__ __volatile__ (					\
	    "leal %1, %%eax \n"					\
	    "addl $4, %%eax \n"					\
	    "movl %%eax, %0 \n" :: "g" (va), "g" (last));	\
} while (0)

#define plva_end(va)

#else
#error "CRAPPY STDARG NOT DEFINED FOR TARGET ARCHITECTURE"
#endif	/* PLARCH_i386 */


/**
 * Peludo Kernel Termination.
 */
void
krelease(long);


/**
 * Initialize the PLC api.
 * Initialize the system dependent part of PLC, set the errno locator
 * to point to the unthreaded errno variable, the h_errno locator to
 * point to the unthreaded h_errno variable, and the cancellation
 * callbacks, initialize the memory manager, peludo sockets and the
 * peludo implementation of the C standard api.
 * <p>
 * This function is automatically called during initialization.
 * </p>
 *
 * @return A Peludo Status code is returned.
 *
 * @see PLSTATUS
 * @see pl_unthreaded_errno()
 * @see pl_unthreaded_h_errno()
 * @see plrelease()
 * @see plsock_init()
 * @see plsys_init()
 */
PLSTATUS
plinit();


/**
 * Release the PLC api.
 * Release the socket subsystem, the memory manager and the
 * low level PLC api.
 * <br>
 * This function should only be called by the user at termination
 * time when the PLC api is statically linked to the application.
 * <br>
 * You should not care of this function when your application is
 * dynamically linked to the PLC api.
 *
 * @see plinit()
 * @see plmm_release()
 * @see plsock_release()
 * @see plsys_release()
 */
void
plrelease();


/**
 * Call the exit callback and terminate the process.
 * If the user has defined its own exit callback it is called;
 * if that function returns or if it is not defined the process
 * is terminated by calling plsys_exit().
 *
 * @param retval (IN) Exit value.
 *
 * @see plsetexit()
 * @see plsys_exit()
 */
void
plexit(long retval);


/**
 * Set an exit callback.
 *
 * @param fn (IN) Exit callback to set;
 *
 * @param old (OUT) If not NULL updated with the previous callback.
 *
 * @see _exit()
 * @see plexit()
 */
void
plsetexit(void (*fn)(long), void (**old)(long));


/**
 * Set errno function locator.
 * You don't want to use this function unless you're coding
 * a threading library.
 *
 * @param errnofn (IN) Pointer to the function that returns a
 * pointer to the errno variable.
 *
 * @see plgeterrno()
 */
void
plseterrno(int *(*errnofn)(void));


/**
 * Errno locator.
 * You want to use the errno define instead of being calling this
 * function directly.
 *
 * @return A pointer to the calling thread's errno variable is returned.
 *
 * @see errno
 * @see plseterrno()
 */
extern int *
(*plgeterrno)();


/**
 * @def errno
 * This is the peludo version of the C standard errno(3) variable.
 *
 * @see plgeterrno()
 */
#define errno	(*plgeterrno())


/**
 * Set errno function locator.
 * You don't want to use this function unless you're coding
 * a threading library.
 *
 * @param herrnofn (IN) Pointer to the function that returns a
 * pointer to the h_errno variable.
 *
 * @see plgetherrno()
 */
void
plsetherrno(int *(*herrnofn)(void));


/**
 * h_errno locator.
 * You want to use the h_errno define instead of being calling this
 * function directly.
 *
 * @return A pointer to the calling thread's h_errno variable is returned.
 *
 * @see h_errno
 * @see plsetherrno()
 */
extern int *
(*plgetherrno)();


/**
 * @def h_errno
 * This is the peludo version of networking h_errno variable.
 *
 * @see plgetherrno()
 */
#define h_errno	(*plgetherrno())


/** 
 * Compare memory.
 * This is the Peludo version of the C standard memcmp(3).
 */
int
plmemcmp(const void *, const void *, size_t);


/**
 * Copies a buffer into another (buffers may overlap).
 * @param dst (IN) Destination buffer;
 * @param src (IN) Source buffer;
 * @param size (IN) Number of bytes to copy.
 * @return A pointer to the destination buffer.
 */
void *
plsys_memmove(void *dst, const void *src, size_t size);


/**
 * Fill memory with a specified value.
 * @param dst (IN) Destination buffer;
 * @param byte (IN) Value;
 * @param size (IN) Number of bytes to fill.
 * @return A pointer to the destination buffer.
 */
void *
plsys_memset(void *dst, uint8 byte, size_t size);


/**
 * Concatenate two strings.
 * This is the Peludo version of the C standard strcat(3).
 */
char *
plstrcat(char *, const char *);


/**
 * Safely concatenate two strings.
 * This is the Peludo version of the BSD standard strlcat(3).
 * <br>
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left). At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
 *
 * @return Returns strlen(src) + MIN(siz, strlen(initial dst)).
 * If retval >= siz, truncation occurred.
 *
 * @author OpenBSD Project
 */
size_t
plstrlcat(char *dst, const char *src, size_t siz);


/**
 * Compare two strings.
 * This is the Peludo version of the C standard strcmp(3).
 */
int
plstrcmp(const char *, const char *, size_t);


/**
 * Copy a string.
 * This is the Peludo version of the C standard strcpy(3).
 */
char *
plstrcpy(char *, const char *);


/**
 * Copy a string.
 * This is the Peludo version of the C standard strncpy(3).
 */
char *
plstrncpy(char *, const char *, size_t);


/**
 * Return the lenght of a given string.
 * This is the Peludo version of the C standard strlen(3).
 */
size_t
plstrlen(const char *);


/**
 * Compare an unicode string against an asciiz.
 *
 * @param uni (IN) Unicode string;
 *
 * @param s2 (IN) Asciiz string;
 *
 * @return < 0 if uni < s1; > 0 when uni > s2; and 0 when uni == s2.
 *
 * @bug This function works only when the unicode string contains
 * embedded ascii or latin-1 characters.
 */
int
plunistrcasecmp(const char *uni, const char *s2);


/**
 * Write a formatted string to standard output.
 *
 * @param fmt (IN) String containing the output format specification
 * (and how to treat the following arguments, if any);
 *
 * @return The number of written bytes is returned.
 *
 * @see plvprint()
 */
int
plprint(const char *fmt, ...);


/**
 * Write a formatted string to standard output.
 *
 * @param fmt (IN) String containing the output format specification
 * (and how to treat the following argument);
 *
 * @param args (IN) List of arguments.
 *
 * @return The number of written bytes is returned.
 *
 * @see plputs()
 * @see plvnprint_cb()
 */
int
plvprint(const char *fmt, plva_list args);


/**
 * Write a formatted string to a specified memory buffer.
 *
 * @param dst (IN) Buffer where to write the formatted string;
 *
 * @param fmt (IN) String containing the output format specification
 * (and how to treat the following arguments, if any);
 *
 * @return The number of written bytes is returned.
 *
 * @see plsnprint()
 * @see plvsprint()
 *
 * @bug This function does not check the destination buffer lenght. Use
 * plsnprint() instead.
 */
int
plsprint(char *dst, const char *fmt, ...);


/**
 * Write a formatted string to a specified memory buffer.
 *
 * @param dst (IN) Buffer where to write the formatted string;
 *
 * @param fmt (IN) String containing the output format specification
 * (and how to treat the following argument);
 *
 * @param args (IN) List of arguments.
 *
 * @return The number of written bytes is returned.
 *
 * @see plvnprint_cb()
 * @see plvsnprint()
 *
 * @bug This function does not check the destination buffer lenght. Use
 * plvsnprint() instead.
 */
int
plvsprint(char *dst, const char *fmt, plva_list args);


/**
 * Write a formatted string to a specified memory buffer.
 *
 * @param dst (IN) Buffer where to write the formatted string;
 *
 * @param siz (IN) Size of the destination buffer;
 *
 * @param fmt (IN) String containing the output format specification
 * (and how to treat the following arguments, if any);
 *
 * @return The number of written bytes is returned.
 *
 * @see plvsnprint()
 */
int
plsnprint(char *dst, size_t siz, const char *fmt, ...);


/**
 * Write a formatted string to a specified memory buffer.
 *
 * @param dst (IN) Buffer where to write the formatted string;
 *
 * @param siz (IN) Size of the destination buffer;
 *
 * @param fmt (IN) String containing the output format specification
 * (and how to treat the following argument);
 *
 * @param args (IN) List of arguments.
 *
 * @return The number of written bytes is returned.
 *
 * @see plvnprint_cb()
 */
int
plvsnprint(char *dst, size_t siz, const char *fmt, plva_list args);


/**
 * Send an asciiz to standard ouptut.
 * Unlinke the C standard puts(3) this function does not add any
 * trailing newline character.
 *
 * @param str (IN) String to write print.
 *
 * @return The number fo written bytes is returned.
 *
 * FIXME: Move to sysdep? What is stdout in windows?
 */
int
plputs(const char *str);


/**
 * Get the string description for a given Peludo Status code.
 *
 * @param status (IN) Peludo Status code;
 *
 * @return When compiled in debug mode a static string containing
 * the description of the Peludo Status code is returned; when compiled
 * in release mode an empty static string is returned.
 *
 * @see PLSTATUS
 */
const char *
plstrstatus(PLSTATUS status);


/**
 * Zeroe memory.
 * @see plsys_memset(void *, uint8, size_t)
 */
#define	plbzero(s, n)	plsys_memset((s), 0, (n))


/**
 * Return the lower case of a given ascii character.
 */
#define pltolow(a)	((a) | 0x20)


/**
 * Return the upper case of a given ascii character.
 */
#define pltoupp(a)	((a) & ~0x20)


void plsetcancel(void (*)(), void (*)(), void (**)(), void (**)());

extern void (*plenter_cancel_point)();
extern void (*plleave_cancel_point)();

PLSTATUS plmm_init();
void	 plmm_release();
PLSTATUS plmm_realloc(long *, long *, size_t);
PLSTATUS plmm_free(long *, long);


/**
 * Alloc a new block of memory.
 *
 * @param err (OUT) If not NULL, on error it is updated with the
 * system error code (think of it as errno);
 *
 * @param addr (OUT) On success it is updated with the address of the
 * allocated memory block;
 *
 * @param siz (IN) Size (in bytes) of the memory block to allocate.
 *
 * @return A Peludo Status code is returned.
 *
 * @see plmm_realloc()
 */
static inline PLSTATUS __attribute__((always_inline))
plmm_malloc(long *err, long *addr, size_t siz)
{
	*addr = 0;
	return plmm_realloc(err, addr, siz);
}


/**
 * @def plmalloc(siz)
 * Allocate a new block of memory.<br>
 * This is the Peludo implementation of the C standard malloc(3).
 *
 * @param siz (IN) Size of the block of memory to allocate.
 *
 * @return On success a pointer to the allocated block of memory
 * is returned; on error the errno variable is updated and NULL is
 * returned.
 *
 * @see plrealloc()
 */
#define plmalloc(siz)	plrealloc(NULL, (siz))


/**
 * [Re]alloc a block of memory.
 * This is the Peludo implementation of the C standard realloc(3).
 * <br>
 * This function uses the Peludo Simple Memory Manager api.
 *
 * @param ptr (IN) Address of the memory block to realloc or NULL;
 *
 * @param size (IN) Size (in bytes) of the memory block to [re]alloc.
 *
 * @return On success a pointer to the [re]allocated block of memory
 * is returned; on error the errno variable is updated and NULL is
 * returned.
 *
 * @see plfree()
 * @see plmm_realloc()
 */
void *
plrealloc(void *ptr, size_t size);


/**
 * Release memory allocated using plmalloc().
 * This is the Peludo implementation of the C standard free(3).
 *
 * @param addr (IN) Starting address of the memory block to release.
 *
 * @see plmalloc()
 * @see plmm_free()
 */
void
plfree(void *addr);


/**
 * Block all signals.
 * Block all signals including SIGCANCEL.
 *
 * @param oset (OUT) Updated with the signal mask before the blocking.
 *
 * @return Nothing. On error a debug message is printed and the system is
 * halted.
 *
 * @see plsigunblock()
 */
void
plsigblock(sigset_t *oset);


/**
 * Set the specified signal mask.
 *
 * @param cset (IN) Signal mask to set (this mask should be obtained by
 * calling plsigblock()).
 *
 * @return Nothing. On error a debug message is printed and the system is
 * halted.
 *
 * @see plsigblock()
 */
void
plsigunblock(const sigset_t *cset);


/**
 * Native OS syscall.
 * This function delivers a native OS syscall; in the case of failure
 * the native error code is returned as a negative value.
 * <br>
 * This function is commonly used to build the low level native component
 * of the PLC api. You don't want to use this function directly.
 *
 * @param err (IN) In the case of error this variable is updated with
 * the native error code;
 *
 * @param sysno (IN) Syscall Number; the syscall number defines all
 * the following arguments.
 *
 * On linux-i386 after the sysno parameter follows 6 arguments no matter
 * whether the specified syscall requires them or not.
 *
 * @return On success the return value is system dependent;
 * on error a negative value is returned.
 */
long
_plsyscall(long *err, long sysno, ...);


/**
 * Native OS syscall.
 * This function delivers a native OS syscall; in the case of failure
 * the native error code is placed into the standard errno "variable".
 * <br>
 * This function is commonly used to build the low level native component
 * of the PLC api. You don't want to use this function directly.
 *
 * @param sysno (IN) Syscall Number; the syscall number defines all
 * the following arguments.
 *
 * On linux-i386 after the sysno parameter follows 6 arguments no matter
 * whether the specified syscall requires them or not.
 *
 * @see plgeterrno()
 *
 * @return On success the return value is system dependent;
 * on error a negative value is returned.
 */
long
plsyscall(long sysno, ...);


/**
 * Initialize the PLC low level system.
 * (all functions starting with plsys_xxx()).
 * This function is called by plinit() during the initialization
 * of the PLC API.
 *
 * @return A Peludo Status code is returned.
 *
 * @see plinit()
 */
PLSTATUS
plsys_init();


/**
 * Release the PLC low level system.
 * This function is ccalled by plrelease() during uninitialization
 * of the PLC API.
 *
 * @see plrelease()
 */
void
plsys_release();


/**
 * @return The number of available processor is returned;
 *
 * @bug Returns 1 always.
 */
int
plsys_getnprocs();


/**
 * Relinquish the processor.
 *
 * @return A Peludo Status code is returned.
 *
 * @see PLSTATUS
 */
PLSTATUS
plsys_schedyield();


/**
 * @return The current process Id is returned.
 */
long
plsys_getpid();


/**
 * @return The current thread Id is returned.
 */
long
plsys_gettid();


/**
 * Fork a new process.
 *
 * @param pid (IN) On success it is updated with the new process id
 * (0 means parent process);
 *
 * @return A Peludo status code.
 */
PLSTATUS
plsys_fork(long *pid);

pid_t plsys_setsid();
int plsys_execve(const char *, char *const[], char *const[]);


/**
 * Terminate the calling process and return to the system.
 *
 * @param val (IN) Exit value.
 */
void
plsys_exit(long val);


/**
 * Send a signal to a specified process.
 *
 * @param pid (IN) Process ID;
 *
 * @param signo (IN) Signal number.
 *
 * @return A Peludo Status code is returned.
 */
PLSTATUS
plsys_kill(long tid, long signo);


/**
 * Terminate the calling thread.
 *
 * @param val (IN) Exit value.
 */
void
plsys_texit(long val);


/**
 * Fork a new thread.
 *
 * @param tid (OUT) If not NULL it is updated with the new thread Id;
 *
 * @param stack (IN) Stack for the new thread;
 *
 * @param fn (IN) Function to be executed by the new thread;
 *
 * @param arg (IN) Argument to pass to the function.
 *
 * @return The parent thread receives a Peludo Status code and the
 * new thread id into the tid parameter; the child thread is spawned
 * executing the fn() function, if that function returns the child is
 * automatically termiated through plsys_texit() with exit code 0.
 *
 * @see PLSTATUS
 */
PLSTATUS
plsys_tfork(long *tid, long stack, void (*fn)(long), long arg);


/**
 * Send a signal to a specified thread.
 *
 * @param tid (IN) Thread ID;
 *
 * @param signo (IN) Signal number.
 *
 * @return A Peludo Status code is returned.
 *
 * @see PLSTATUS
 */
PLSTATUS
plsys_tkill(long tid, long signo);


/**
 * Put the current thread to sleep if a given memory address
 * contains a specified value.
 *
 * @param addr (IN) Memory address to test;
 *
 * @param val (IN) Value to test against;
 *
 * @param timeout (IN) Timeout in milliseconds (-1 means no timeout).
 *
 * @return Peludo Status code:
 * - If was woken by another thread PLSTATUS_SUCCESS is returned;
 * - If was woken by a timeout PLSTATUS_TIMEOUT is returned;
 * - If both values are not equal PLSTATUS_TRY_AGAIN is returned;
 * - On error PLSTATUS_FAILED is returned (the error code is into errno).
 * - If addr is NULL PLSTATUS_INVALID is returned.
 *
 * @see PLSTATUS
 * @see plsys_tsignal()
 */
PLSTATUS
plsys_twait(void *addr, long val, long timeout);


/**
 * Wakeup one or all threads sleeping on a specified memory address.
 *
 * @param addr (IN) Memory address;
 *
 * @param wall (IN) A non-zero value will wakeup all threds sleeping
 * on the specified address; a 0 will wakeup only one thread.
 *
 * @return The number of threads woken up is returned.
 *
 * @see plsys_twait()
 */
long
plsys_tsignal(void *addr, int wall);


/**
 * Put the calling thread to sleep until a signal is received.
 * If the thread receives a SIGCANCEL it is terminated using plexit().
 *
 * @param set (IN) Signal set specifying the signals to be waited for;
 *
 * @param signo (OUT) If not NULL, updated with the signal that woke
 * up the thread.
 *
 * @return A Peludo Status code is returned.
 *
 * NOTE from linux manpage:
 * For sigwait to work reliably, the signals being waited for must be
 * blocked in all threads, not only in the calling thread, since otherwise
 * the POSIX semantics for signal delivery do not guarantee that it’s the
 * thread doing the sigwait that will receive the signal. The best way
 * to achieve this is block those signals before any threads are created,
 * and never unblock them in the program other than by calling sigwait.
 *
 * @see PLSTATUS
 * @see sigset_t
 * @see plexit()
 */
PLSTATUS
plsys_tsigwait(const sigset_t *set, int *signo);


/**
 * Set/Get the signal mask for the calling thread.
 * This is the peludo implementation of POSIX sigprocmask(3) and
 * pthread_sigmask(3).
 *
 * @param how (IN) How the signal newset should be treated;
 *
 * @param newset (IN) New signal set;
 *
 * @param oldset (OUT) Old (current) signal set.
 *
 * @return A peludo status code is returned.
 *
 * @see PLSTATUS
 * @see sigset_t
 * @see _plsys_tsigmask()
 */
PLSTATUS
plsys_tsigmask(int how, const sigset_t *newset, sigset_t *oldset);


/**
 * Set/Get the signal mask for the calling thread.
 * This function is almost a direct call to the native sigprocmask(2)
 * syscall  and is intended to be used by a threading library; most
 * probably you want to use plsys_tsigmask() instead.
 *
 * @return A peludo status code is returned.
 *
 * @see PLSTATUS
 * @see sigset_t
 * @see plsys_tsigmask()
 *
 * @todo What about windows?
 */
PLSTATUS
plsys_tsigmask_nt(int how, const sigset_t *newset, sigset_t *oldset);


PLSTATUS
plsys_sigaction(int, const struct sigaction *, struct sigaction *);


/**
 * Suspend the current process.
 * This function is the peludo implementation of POSIX sigsuspend().
 * <br>
 * Suspend the current process until the reception of a signal not blocked
 * by the specified mask.
 *
 * @param mask (IN) Mask to set atomically when suspending.
 *
 * @return PLSTATUS_FAILED is returned.
 *
 * @see sigset_t
 */
PLSTATUS
plsys_sigsuspend(const sigset_t *mask);


/**
 * Allocate pages of system memory.
 *
 * @param err (OUT) On error it is updated with the native error code;
 *
 * @param addr (OUT) Updated with the start address of allocated pages;
 *
 * @param size (IN) Size to alloc (this value is automatically aligned
 * to the system's page size);
 *
 * @param flags (IN) Protection flags.
 *
 * @return A Peludo Status code is returned; if the returned code is
 * PLSTATUS_FAILED the err argument is updated with the native error
 * code.
 *
 * @see PLMFLAGS
 * @see PLSTATUS
 * @see _plsys_mprotect()
 * @see _plsys_pfree()
 */
PLSTATUS
_plsys_palloc(long *err, long *addr, size_t size, PLMFLAGS flags);


/**
 * Release pages of system memory.
 *
 * @param err (OUT) On error it is updated with the native error code;
 *
 * @param addr (IN) Starting address of pages to release;
 *
 * @param size (IN) Size of memory to release (this value is automatically
 * aligned to the system's page size).
 *
 * @return A Peludo Status code is returned; if the returned code is
 * PLSTATUS_FAILED the err argument is updated with the native error code.
 *
 * @see PLSTATUS
 * @see _plsys_palloc()
 */
PLSTATUS
_plsys_pfree(long *err, long addr, size_t size);


/**
 * Set the protection flags of specified memory pages.
 *
 * @param err (OUT) On error updated with the native error code;
 *
 * @param addr (IN) Starting address to protect (this value should be
 * page aligned);
 *
 * @param size (IN) Size of memory to protect (this value is automatically
 * aligned to the system's page size);
 *
 * @param flags (IN) Protection flags.
 *
 * @return A Peludo Status code is returned; if the returned code is
 * PLSTATUS_FAILED the err argument is updated with the system error code.
 *
 * @see PLMFLAGS
 * @see PLSTATUS
 * @see _plsys_mprotect()
 * @see _plsys_palloc()
 */
PLSTATUS
_plsys_mprotect(long *err, long addr, size_t size, PLMFLAGS flags);


/**
 * Allocate pages of system memory.
 *
 * @param addr (OUT) Updated with the start address of allocated pages;
 *
 * @param size (IN) Size to alloc (this value is automatically aligned
 * to the system's page size);
 *
 * @param flags (IN) Protection flags.
 *
 * @return A Peludo Status code is returned; if the returned code is
 * PLSTATUS_FAILED the system error is set into the global errno variable.
 *
 * @see PLMFLAGS
 * @see PLSTATUS
 * @see plsys_mprotect()
 * @see plsys_pfree()
 */
PLSTATUS
plsys_palloc(long *addr, size_t size, PLMFLAGS flags);


/**
 * Release pages of system memory.
 *
 * @param addr (IN) Starting address of pages to release;
 *
 * @param size (IN) Size of memory to release (this value is automatically
 * aligned to the system's apge size).
 *
 * @return A Peludo Status code is returned; if PLSTATUS_FAILED is returned
 * the global errno variable is updated with the system error code.
 *
 * @see PLSTATUS
 * @see plsys_palloc()
 */
PLSTATUS
plsys_pfree(long addr, size_t size);


/**
 * Set the protection flags of specified memory pages.
 *
 * @param addr (IN) Starting address of to protect (this value should be
 * page aligned);
 *
 * @param size (IN) Size of memory to protect (this value is automatically
 * aligned to the system page size);
 *
 * @param flags (IN) Protection flags.
 *
 * @return A Peludo Status code is returned; if the returned code is
 * PLSTATUS_FAILED the system error is set into the global errno variable.
 *
 * @see PLMFLAGS
 * @see PLSTATUS
 * @see plsys_palloc()
 */
PLSTATUS
plsys_mprotect(long addr, size_t size, PLMFLAGS flags);


/**
 * Convert from Peludo Memory Protection flags to native potection flags.
 *
 * @param flags (IN) Peludo Memory Protection Flags.
 *
 * @return A system dependent memory protection flags is returned.
 *
 * @see PLMFLAGS
 */
int
plsys_plmflags2prot(PLMFLAGS flags);


PLSTATUS plsys_access(const char *, long);
PLSTATUS plsys_open(long *, const char *, long, long);
PLSTATUS plsys_close(long);
PLSTATUS plsys_dup2(long, long);
PLSTATUS plsys_fcntl(long *, long, long, long);


/**
 * Get the current working directory.
 *
 * @param dst (OUT) Destination buffer to fill with the current
 * directory pathname;
 *
 * @param dstsiz (IN) Size of the destination buffer.
 *
 * @return A Peludo Status code.
 *
 * @see PLSTATUS
 */
PLSTATUS plsys_getcwd(char *dst, long dstsiz);


/**
 * Rename a file.
 *
 * @param oldpath (IN) Old pathname;
 *
 * @param newpath (IN) New pathname.
 *
 * @return A Peludo Status code.
 *
 * @see PLSTATUS
 */
PLSTATUS plsys_rename(const char *oldpath, const char *newpath);


/**
 * Read data from an open file.
 *
 * @param fd (IN) File descriptor or handle to the open file;
 *
 * @param buf (OUT) Buffer where to put read data;
 *
 * @param size (IN) Number of bytes to read;
 *
 * @param rsiz (OUT) If not NULL on success it is updated with the
 * number of read bytes.
 *
 * @return A Peludo Status code is returned.
 *
 * @see PLSTATUS
 * @see plsys_close()
 * @see plsys_open()
 * @see plsys_write()
 */
PLSTATUS
plsys_read(long fd, void *buf, size_t size, size_t *rsiz);


/**
 * Write a buffer to the specified file.
 *
 * @param fd (IN) File descriptor or handle to the open file;
 *
 * @param buf (IN) Buffer to write;
 *
 * @param size (IN) Number of bytes to write;
 *
 * @param wsiz (OUT) If not null on success it is updated with the
 * number of bytes written.
 *
 * @return A Peludo Status code is returned.
 *
 * @see PLSTATUS
 * @see plsys_close()
 * @see plsys_open()
 * @see plsys_read()
 */
PLSTATUS
plsys_write(long fd, const void *buf, size_t size, size_t *wsiz);


PLSTATUS plsys_lseek(long *, long, long, int);
PLSTATUS plsys_lstat(const char *, stat_t *);


/**
 * Get information about an open file.
 *
 * @param fd (IN) File descriptor (handler) to the open file;
 *
 * @param st (OUT) Updated with the file information.
 *
 * @return A Peludo Status code is returned.
 *
 * @see stat_t
 */
PLSTATUS
plsys_fstat(long fd, stat_t *st);


PLSTATUS plsys_fdatasync(long);
PLSTATUS plsys_fsync(long);
PLSTATUS plsys_ftruncate(long, long);
PLSTATUS plsys_stat(const char *, stat_t *);
PLSTATUS plsys_chmod(const char *, long);


int	plsys_chdir(const char *);
pid_t	plsys_waitpid(pid_t, int *, int);


/**
 * Open a directory.
 *
 * @param handle (OUT) Updated with the handle to the opened direcotry;
 *
 * @param path (IN) Pathname.
 *
 * @return A Peludo Status code.
 *
 * @see plsys_closedir()
 * @see plsys_readdir()
 */
PLSTATUS plsys_opendir(long *handle, const char *path);


/**
 * Close a directory.
 * Once closed, the handle cannot be reused.
 *
 * @param handle (IN) Handle to the open directory.
 *
 * @see plsys_opendir()
 * @see plsys_readdir()
 */
void plsys_closedir(long handle);


/**
 * Read a directory.
 *
 * @param handle (IN) Handle to the open directory;
 *
 * @param de (OUT) Updated with the contents of the next directory entry;
 *
 * @return A Peludo Status code. When there are not more entries
 * <code>PLSTATUS_EOF</code> is returned.
 *
 * @see plsys_opendir()
 * @see plsys_closedir()
 */
PLSTATUS plsys_readdir(long handle, struct dirent **de);


/* XXX fix flags and prot */
PLSTATUS
plsys_mmapfile(long *addr, long fd, size_t size, long flags, long prot,
    long offset);

PLSTATUS plsys_mmapflush(long, size_t);
PLSTATUS plsys_unlink(const char *);

PLSTATUS plsys_mkdir(const char *, long);
PLSTATUS plsys_rmdir(const char *);

PLSTATUS plsys_readlink(const char *, char *, long *);


/**
 * Peludo version of C standard gettimeofday(3)
 *
 * @return A Peludo Status code is returned.
 *
 * @see timeval
 * @see timezone
 *
 * XXX should be public?
 */
PLSTATUS
plsys_gettimeofday(struct timeval *tv, struct timezone *tz);


/**
 * Convert from miliseconds to timespec.
 *
 * @param tv (OUT) If not NULL, updated with the converted time;
 *
 * @param ms (IN) Miliseconds.
 *
 * @see timespec
 */
void
plsys_ms2timespec(struct timespec *tv, long ms);


/**
 * Get system time in seconds.
 *
 * @param time (OUT) Updated with the system time.
 *
 * @return A Peludo Status code is returned.
 */
PLSTATUS
plsys_time(time_t *time);


/**
 * Put the current thread to sleep.
 *
 * @param ms (IN) Number of milliseconds to sleep.
 */
void plsys_msleep(long ms);


/**
 * Deliver a ioctl.
 *
 * @param fd (IN) File descriptor or handle;
 *
 * @param req (IN) Request code;
 *
 * @param arg (IN) Argument;
 *
 * @return A Peludo Status code is returned.
 *
 * @todo Make it system independent or move to plsys_xx().
 */
PLSTATUS plsys_ioctl(long fd, int req, void *arg);


PLSTATUS plsys_uname(struct utsname *);
PLSTATUS plsys_gethostname(char *, size_t);
PLSTATUS plsys_pipe(long *);
PLSTATUS plsys_pipe(long *);

PLSTATUS plsys_poll(long *, struct pollfd *, nfds_t, int);

PLSTATUS plsys_sigaddset(sigset_t *, int);
PLSTATUS plsys_sigdelset(sigset_t *, int);
PLSTATUS plsys_sigemptyset(sigset_t *);
PLSTATUS plsys_sigfillset(sigset_t *);
int	 plsys_sigismember(const sigset_t *, int);


/**
 * Save current execution state.
 * Think of this function as the peludo version of the BSD standard
 * _setjmp() but don't rely on that, its behaviour is a bit different.
 * It is not fully compatible.
 *
 * @param buf (OUT) Buffer where to put current environment state;
 *
 * @see jmp_buf
 *
 * @todo This function is not complete; the current implementation
 * just saves some processor registers into the jmp_buf.
 * XXX
 */
int
plsys__setjmp(jmp_buf *env);


/**
 * Wait until some file descriptors or windows socket become ready.
 *
 * @param ready (OUT) Updated with the number of ready sockets;
 *
 * @param fds (IN) Highest numbered file descriptor in any of the sets
 * (under windows this argument is ignored);
 *
 * @param rfd (IN/OUT) Set sockets to be checked for readability;
 *
 * @param wfd (IN/OUT) Set of sockets to be checked for writability;
 *
 * @param efd (IN/OUT) Set of sockets to be checked for exceptions;
 *
 * @param tout (IN) Timeout.
 *
 * @return A Peludo Status code is returned. On timeout the 'ready'
 * parameter is set to 0 and the returned code is PLSTATUS_TIMEOUT.
 *
 * @see PLSTATUS
 *
 * FIXME move to plsock_select()? On windows this works only with sockets.
 * XXX remove, select() should be deprecated on any system.
 */
PLSTATUS
plsys_select(int *ready, int fds, fd_set *rfd, fd_set *wfd, fd_set *efd,
    struct timeval *tout);


/* XXX not portable */
PLSTATUS PLEXPORT  plsys_munmap(long, size_t);


/* XXX */
long plsys_getdll(const char *);
long plsys_getsymbol(long, const char *);
long plsys_getsymbol_any(const char *);


/**
 * Load a Native DLL.
 * If the desired DLL is not mapped into the address space of the
 * current process it is loaded using a builtin ELF loader. On windows
 * it is loaded using the native LoadLibraryA() function.
 *
 * @param dll (OUT) Updated with the base address of the loaded DLL;
 *
 * @param dllname (IN) Asciiz string containing DLL's name.
 *
 * @return A Peludo Status code is returned.
 *
 * @see PLSTATUS
 *
 * @todo Not implemented on unix.
 */              
PLSTATUS
plsys_dlopen(long *dll, const char *dllname);


/**
 * Get the address of a symbol exported by a DLL.
 *
 * @param addr (OUT) Updated with the address of the desired symbol;
 *
 * @param dll (IN) Base address of the DLL (if it is 0 this function
 * traverses all loaded DLLs searching for the symbol);
 *
 * @param symname (IN) Asciiz string containing the symbol name.
 *
 * @return A Peludo Status code is returned.
 *
 * @see PLSTATUS
 * @see plsys_dlopen()
 * @see plsys_getsymbol()
 * @see plsys_getsymbol_any()
 *
 * @todo Not implemented on unix.
 */
PLSTATUS
plsys_dlsym(long *addr, long dll, const char *symname);


/**
 * Entry point when compiled as a native executable.
 * This function is called by the OS native loader when all
 * modules are statically linked into a single ELF or PE file.
 * <br>
 * This function is part of the peludo crt0.o file.
 * <br>
 * It calls the main() function and terminates the process
 * returning to the system with its output value.
 * <br>
 * You should not worry about this function: The Peludo compilation
 * system put everything in the correct place.
 *
 * @see plsys_exit()
 */
//void _start();


/**
 * Entry point when compiled in PLD format.
 * This function is called by the PLD loader (a Peludo Kernel component)
 * when a .plb file is launched either via prun or executed directly
 * from memory (in both cases all peludo modules are "linked" in PLD
 * format).
 * <br>
 * This function is part of the peludo plcrt0.o file.
 * <br>
 * It calls the module's main() function and returns to the caller
 * (the PLD loader) with its output value.
 * This function also sets krelease() as the termination callback in
 * order to return to the peludo kernel when the end of main() is
 * never reached (for example when the process calls plexit()).
 * <br>
 * You should not worry about this function: The Peludo compilation
 * system put everything in the correct place.
 *
 * @see krelease()
 * @see plexit()
 * @see plsetexit()
 */
/*
int
_plstart(int argc, char **argv);
*/


/**
 * @function plsys_halt()
 * Abort the current process execution.
 *
 * XXX move plsys_halt() sysdep and exit nicely with an "abort" message
 */
#if defined(PLARCH_i386)
static inline void __attribute__((always_inline))
plsys_halt()
{
	//XXX plsys_exit(-1);
	asm("hlt");
        /* NOTREACHED */
}

#else
#error "plsys_halt() NOT DEFINED FOR TARGET ARCHITECTURE"
#endif	/* PLARCH_i386 */


/**
 * @function plsys_curpage()
 * @return Return the base address of current executing page.
 */
#if defined(PLARCH_i386)
static inline long __attribute__((always_inline))
plsys_curpage()
{
	long ret;

	__asm__ __volatile__(
	    "		call	_l1			\n"
	    "_l1:	popl    %%eax			\n"
	    "		andl    $0xfffff000, %%eax	\n"
	    "		movl	%%eax, %0		\n"
	    :: "g" (ret));

	return ret;
}

#else
#error "plsys_curpage() NOT DEFINED FOR TARGET ARCHITECTURE"
#endif	/* PLARCH_i386 */


/**
 * @define PLTRAP()
 * Generate a trap interrupt.
 */
#if defined(PLARCH_i386)
#define PLTRAP()	__asm__ __volatile__("int3\n")
#else
#error "PLTRAP NOT DEFINED FOR TARGET ARCHITECTURE"
#endif	/* PLARCH_i386 */


#endif	/* PLC_H */
