/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/include/plmutex.h
 * Peludo Mutex.
 * FIXME: used?
 */
#ifndef PLMUTEX_H
#define PLMUTEX_H

#include <plarchdef.h>
#include <pldef.h>


/**
 * Peludo Mutex.
 */
typedef volatile struct _PLMUTEX {
	long __attribute__((aligned(sizeof(long)))) mtx_lock;
	long __attribute__((aligned(sizeof(long)))) mtx_count;
} PLMUTEX;

#define PLMUTEX_INITIALIZED		{ 0, 0 }


/**
 * Initialize a Peludo Mutex.
 * The mutex is initialized unlocked.
 *
 * @param mutex (IN) Peludo Mutex to initialize.
 *
 * @see PLMUTEX
 * @see plmutex_lock()
 * @see plmutex_unlock()
 */
void PLEXPORT
plmutex_init(PLMUTEX *mutex);


/**
 * Lock a Peludo Mutex.
 * Peludo Mutexes are recursive; when a thread is locking an
 * already owned mutex an internal counter is incremented;
 * To completely unlock that mutex an equal number of plmutex_unlock()
 * operations must be done.
 * <br>
 * <b>XXX Peludo mutexes are not signal safe: A signal handler should
 * never call any mutex handling function; otherwise there could be a
 * deadlock.</b>
 *
 * @param mutex (IN) Mutex to lock;
 *
 * @param block (IN) If non-zero the current thread will
 * block until the mutex is owned; otherwise it will just
 * try to own the mutex and exit.
 *
 * @return If the calling thread owned the mutex PLSTATUS_SUCCESS
 * is returned; if the calling thread was not able to own the mutex
 * PLSTATUS_TRY_AGAIN is returned; On error any other Peludo Status
 * code is returned.
 *
 * @see PLMUTEX
 * @see PLSTATUS
 * @see plmutex_init()
 * @see plmutex_siglock()
 * @see plmutex_unlock()
 * @see plsys_twait()
 */
PLSTATUS PLEXPORT
plmutex_lock(PLMUTEX *mutex, int block);


/**
 * Lock a Peludo Mutex.
 * This function is like a blocking plmutex_lock() but signals
 * (including SIGCANEL) are blocked before being waiting to get
 * the lock and unblocked once it is owned.
 *
 * @param mutex (IN) Mutex to lock;
 *
 * @return If the calling thread owned the mutex PLSTATUS_SUCCESS
 * is returned; on error any other Peludo Status code is returned.
 *
 * @see PLMUTEX
 * @see PLSTATUS
 * @see plmutex_lock()
 * @see plmutex_unlock()
 * @see plsigblock()
 */
PLSTATUS PLEXPORT
plmutex_siglock(PLMUTEX *mutex);


/**
 * Unlock a Peludo Mutex.
 * If the calling thread have locked the mutex multiple times an
 * internal counter is decreaced and the mutex is effectively
 * unlocked when it reaches 0.
 *
 * @param mutex (IN) Mutex to unlock.
 *
 * @return On success PLSTATUS_SUCCESS is returned;
 * if the calling thread does not own the mutex PLSTATUS_SUCCESS is
 * returned; On error another Peludo Status code is returned.
 *
 * @see PLMUTEX
 * @see plmutex_init()
 * @see plmutex_lock()
 * @see plsys_tsignal()
 */
PLSTATUS PLEXPORT
plmutex_unlock(PLMUTEX *mutex);


/**
 * Return a mutex owner's TID.
 * 
 * @param mutex (IN) Mutex;
 *
 * @return If the mutex has an owner its TID is returned; otherwise 0 is
 * returned.
 *
 * @see PLMUTEX
 */
long PLEXPORT
plmutex_ownertid(PLMUTEX *mutex);


#endif	/* PLMUTEX_H */
