/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/include/plsock.h
 * Peludo sockets.
 */
#ifndef PLSOCK_H
#define PLSOCK_H

#include <plc.h>


/**
 * Initialize the socket system.
 * <p>On unix this function does nothing; on windows the socket library
 * is loaded and all internal dependencies are resolved.</p>
 * <p>This function is called by <code>plinit()</code>.</p>
 * @see plinit()
 * @see plsock_release()
 */
PLSTATUS
plsock_init();


/**
 * Release the socket system.
 * <p>This function is called by <code>plrelease()</code>.</p>
 * <p>You should only call this function at finalization time when statically
 * linked against plc.a and not using any of the plxxx() functions (that is:
 * you're using only plsock_xxx() and maybe plsys_xxx()).</p>
 * @see plrelease()
 * @see plsock_init()
 */
void
plsock_release();
 

/**
 * Create a socket.
 * @param sock Updated with the new socket;
 * @param af Address family;
 * @param type Type of socket;
 * @param proto Protocol;
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_socket(long *sock, int af, int type, int proto);


/**
 * Prepare a socket to listen.
 * @param sock  Socket to be used to listen;
 * @param backlog Maximum length the queue of pending connections;
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_listen(long sock, int backlog);


/**
 * Associate a local address to a socket.
 * @param sock Socket;
 * @param addr Address;
 * @param len Size of addr structure.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_bind(long sock, const struct sockaddr *addr, socklen_t len);


/**
 * Accept a new connection.
 * @param asock Updated a new socket of the accepted connection;
 * @param sock Listening socket;
 * @param addr  Updated with the address of the accepted peer;
 * @param len Size of addr structure; updated with the size of the
 * address of the accepted peer.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_accept(long *asock, long sock, struct sockaddr *addr, socklen_t *len);


/**
 * Connect a socket to a specified address.
 * @param csock Updated with the connected socket;
 * @param sock Socket;
 * @param addr Address to connect to;
 * @param len Size of addr structure.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_connect(long *csock, long sock, const struct sockaddr *addr,
    socklen_t len);


/**
 * Return name of peer connected to a socket.
 * @param sock Connected socket;
 * @param addr Updated with the address of the peer;
 * @param len At input should contain the size of addr structure;
 * At output is updated with the size of the peer's address.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_getpeername(long sock, struct sockaddr *addr, socklen_t *len);


/**
 * Return the name of a socket.
 * @param sock Socket;
 * @param addr Updated with the socket's name;
 * @param len The size of addr structure; updated with the size of the
 * socket name.
 */
PLSTATUS
plsock_getsockname(long sock, struct sockaddr *addr, socklen_t *len);


/**
 * Retrieve a socket option.
 * @param sock Socket;
 * @param level The level at which the option is defined;
 * @param optname Option to retrieve;
 * @param optval Updated with the value of the requested option;
 * @param optlen Updated with the size of the retrieved value.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_getsockopt(long sock, int level, int optname, void *optval,
    socklen_t *optlen);


/**
 * Set a socket option.
 * @param level The level at which the option is defined;
 * @param optname Option to set;
 * @param optval Pointer to the value of the option to set;
 * @param optlen Size of the value to set.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_setsockopt(long sock, int level, int optname, const void *optval,
    socklen_t optlen);


/**
 * Shutdown a connection.
 * @param sock Socket associated to the connection to shutdown;
 * @param how How to shutdown: SHUT_RD, SHUT_WR or SHUT_RDWR.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_shutdown(long sock, int how);


/**
 * Close a socket.
 * <p>On unix you can close a socket using <code>plsys_close()</code>.
 * This is not true for windows where its <code>closesocket()</code> method
 * should be used. If you are coding a portable peludo module you MUST use
 * this method to close a socket. XXX check this statement</p>
 * @param sock Socket to close.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_close(long sock);


/**
 * Send message to a connected socket.
 * <p>On unix you can use <code>plsys_write()</code> to send messages.
 * This is not true for windows where its <code>send()</code> function
 * should be used. If you are coding a portable peludo module you MUST
 * use this method. XXX check this statement</p>
 * @param sent If not <code>NULL</code> it is updated with the number of
 * bytes transferred;
 * @param sock Socket where to send the message;
 * @param buf Buffer containing the message;
 * @parma len Message length;
 * @param flags See posix send() flags.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_send(size_t *sent, long sock, const void *buf, size_t len, int flags);


/**
 * Send a message to a non-connected socket.
 * @param sent If not <code>NULL</code> it is updated with the
 * number kof bytes transferred;
 * @param sock Socket where to send the message;
 * @param buf Buffer containing the message;
 * @parma len Message length;
 * @param flags See posix send() flags;
 * @param to Address where to send the message;
 * @param tolen Size of 'to' address strcuture.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_sendto(size_t *sent, long sock, const void *buf, size_t len, int flags,
    const struct sockaddr *to, socklen_t tolen);


/**
 * Receive a message from a connected socket.
 * @param rec If not <code>NULL</code> it is updated with the
 * number of bytes received;
 * @param sock Connected socket;
 * @param buf Buffer where to put incoming data;
 * @param len Size of buffer;
 * @param flags See posix recv() flags.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_recv(size_t *rec, long sock, void *buf, size_t len, int flags);


/**
 * Receive a message from a connected socket.
 * @param rec If not <code>NULL</code> it is updated with the
 * number of bytes received;
 * @param sock Connected socket;
 * @param buf uffer where to put incoming data;
 * @param len Size of buffer;
 * @param flags See posix recv() flags;
 * @param from pdated with the address of sender;
 * @param fromlen The size of the 'from' buffer; updated with the size
 * of the sender address.
 * @return A Peludo Status code.
 */
PLSTATUS
plsock_recvfrom(size_t *rec, long sock, void *buf, size_t len, int flags,
    struct sockaddr *from, socklen_t *fromlen);


/**
 * Send a message through a connection-mode or connectionless-mode socket.
 * <p>If the socket is connectionless-mode, the message shall be sent to the
 * address specified by msghdr. If the socket is connection-mode, the
 * destination address in msghdr shall be ignored.</o>
 *
 * @param sent If not <code>NULL</code> it is updated with the number
 * of bytes sent;
 * @param sock Socket;
 * @param msg A <code>msghdr</code> structure, containing both the
 * destination address and the buffers for the outgoing message;
 * The length and format of the address depend on the address family of the
 * socket. The <code>msg_flags</code> member is ignored.
 * @param flags Specifies the type of message transmission; may be
 * <code>0</code> or the following flag:
 * <li><code>MSG_EOR</code> Terminate a record (if supported by the
 * protocol);</li>
 * <li><code>MSG_OOB</code> Send out-of-band data on sockets that
 * support out-of-bound data. The significance and semantics of
 * out-of-band data are protocol-specific.</li>
 * <p>
 * The <code>msg_iov</code> and <code>msg_iovlen</code> fields of message
 * specify zero or more buffers containing the data to be sent.
 * <code>msg_iov</code> points to an array of iovec structures;
 * <code>msg_iovlen</code> shall be set to the dimension of this array.
 * In each <code>iovec</code> structure, the <code>iov_base</code> field
 * specifies a storage area and the <code>iov_len</code> field gives its size
 * in bytes. Some of these sizes can be zero. The data from each storage area
 * indicated by <code>msg_iov</code> is sent in turn.</p>
 *
 * @return A peludo status code. Notice that a successful completion does
 * not guarantee delivery of the message.
 *
 * @see https://www.opengroup.org/onlinepubs/000095399/functions/sendmsg.html
 * @see plsock_recvmsg()
 */
PLSTATUS
plsock_sendmsg(long *sent, long sock, const struct msghdr *msg, int flags);


/**
 * Receive a message from a connection-mode or connectionless-mode socket.
 * <p>It is normally used with connectionless-mode sockets because it permits
 * the application to retrieve the source address of received data.</p>
 *
 * @param recv If not <code>NULL</code> it is updated with length
 * of the received message in bytes;
 * @param socket Socket;
 * @param msg Updated with a <code>msghdr</code> structure, containing
 * both the buffer to store the source address and the buffers for the
 * incoming message. The length and format of the address depend on the
 * address family of the socket. The <code>msg_flags</code> member is
 * ignored on input, but may contain meaningful values on output.
 * @parma flags Specifies the type of message reception. Values of
 * this argument are formed by logically OR'ing zero or more of the following
 * values:
 * <li><code>MSG_OOB</code> Requests out-of-band data. The significance
 * and semantics of out-of-band data are protocol-specific.</li>
 * <li><code>MSG_PEEK</code> Peeks at the incoming message.</li>
 * <li><code>MSG_WAITALL</code> On SOCK_STREAM sockets this requests that
 * the function block until the full amount of data can be returned.</li>
 * <p>
 * Upon successful completion, the <code>msg_flags<code> member of the message
 * header shall be the bitwise-inclusive OR of all of the following flags that
 * indicate conditions detected for the received message:
 * <li><code>MSG_EOR</code> End-of-record was received (if supported by the
 * protocol).</li>
 * <li><code>MSG_OOB</code> Out-of-band data was received.</li>
 * <li><code>MSG_TRUNC</code> Normal data was truncated.</li>
 * <li><code>MSG_CTRUNC</code> Control data was truncated. </li>
 * </p>
 *
 * @return A peludo status code.
 *
 * @see https://www.opengroup.org/onlinepubs/000095399/functions/recvmsg.html
 * @see plsock_sendmsg()
 */
PLSTATUS
plsock_recvmsg(long *recv, long sock, struct msghdr *msg, int flags);


/*
 * Create an unbound pair of connected sockets in a specified domain, of a
 * specified type, under the desired protocol.
 * @param domain Domain;
 * @param type Socket type;
 * @param protocol Protocol;
 * @param sv Updated with the created sockets.
 * @return A peludo status code.
 * @see http://www.opengroup.org/onlinepubs/009695399/functions/socketpair.html
 */
PLSTATUS
plsock_socketpair(int domain, int type, int protocol, long *sv);


#endif	/* PLSOCK_H */
