/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/include/platomic.h
 * Peludo Spinlocks.
 */
#ifndef PLSPINLOCK_H
#define PLSPINLOCK_H

#include <platomic.h>


typedef volatile long PLOCK;

#define PLOCK_INITIALIZED		0

#define PLOCK_INIT_UNLOCKED(lock)	(*(lock) = PLOCK_INITIALIZED)


/**
 * @define plock_trylock(PLOCK *lock)
 * Try to get a lock.
 *
 * @param lock <code>PLOCK</code> variable to lock.
 *
 * @return A non-zero value if the lock is gotten; <code>0</code> otherwise.
 *
 * @see PLOCK
 * @see platomic_iex()
 * @see plock_splock()
 */
#define plock_trylock(lock)	(platomic_iex((long *)(lock), 0, 1) == 0)


/**
 * @define plock_splock(PLOCK *lock)
 * Loop until a lock is gotten.
 *
 * @param lock <code>PLOCK</code> variable to lock.
 *
 * @see PLOCK
 * @see platomic_iex()
 * @see plock_trylock()
 * @see plsys_schedyield()
 */
#define plock_splock(lock)	do {				\
	while (platomic_iex((long *)(lock), 0, 1) != 0)		\
		plsys_schedyield();				\
} while (0)


/**
 * @define plock_unlock(PLOCK *lock)
 * Unlock a lock.
 *
 * @param lock <code>PLOCK</code> to unlock.
 *
 * @see platomic_iex()
 * @see plock_splock()
 * @see plock_trylock()
 */
#define plock_unlock(lock)	do {				\
	platomic_iex((long *)(lock), 1, 0);			\
} while (0)


/**
 * @define plock_tryvalock(PLOCK *lock, long value)
 * Try to get a lock.
 *
 * @param lock <code>PLOCK</code> variable to lock;
 *
 * @param value Value to put into the lock.
 *
 * @return A non-zero value if the lock is gotten; <code>0</code> otherwise.
 *
 * @see PLOCK
 * @see platomic_iex()
 * @see plock_spvalock()
 * @see plock_unvalock()
 */
#define plock_tryvalock(lock, value)				\
	(platomic_iex((long *)(lock), 0, (value)) == 0)


/**
 * @define plock_spvalock(PLOCK *lock, long value)
 * Loop until a valued lock is gotten.
 *
 * @param lock <code>PLOCK</code> variable to lock;
 *
 * @param value Value to put into the lock.
 *
 * @see PLOCK
 * @see platomic_iex()
 * @see plock_tryvalock()
 * @see plock_unvalock()
 * @see plsys_schedyield()
 */
#define plock_spvalock(lock, value)	do {			\
	while (platomic_iex((long *)(lock), 0, (value)) != 0)	\
		plsys_schedyield();				\
} while (0)


/**
 * @define plock_unvalock(PLOCK *lock)
 * Unlock a valued lock.
 * <p>If the lock does not have the specified <code>value</code> it means
 * it is not owned by the calling thread and it will not be unlocked.</o>
 *
 * @param lock <code>PLOCK</code> variable to unlock;
 *
 * @param value Value the lock should have to unlock.
 *
 * @see PLOCK
 * @see platomic_iex()
 * @see plock_spvalock()
 * @see plock_tryvalock()
 */
#define plock_unvalock(lock, value)	do {			\
	platomic_iex((long *)(lock), (value), 0);		\
} while (0)


#endif	/* PLSPINLOCK_H */
