/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file include/std/pthread.h
 * Peludo Threads - Posix Threads.
 * @see http://opengroup.org/onlinepubs/007908775/xsh/pthread.h.html
 */
#ifndef __PTHREAD_H__
#define __PTHREAD_H__

#include <plsysdep.h>
#include <pldef.h>
#include <plspinlock.h>
#include <plist.h>

#ifndef	PTHREAD_DEFAULT_STACK_SIZE_NP
#define	PTHREAD_DEFAULT_STACK_SIZE_NP		65536
#endif

#define	PTHREAD_CREATE_DETACHED			0
#define	PTHREAD_CREATE_JOINABLE			1

#define	PTHREAD_CANCELED			((void *)-1)

#define PTHREAD_CANCEL_ENABLE			0
#define PTHREAD_CANCEL_DISABLE			1
#define PTHREAD_CANCEL_DEFERRED			0
#define PTHREAD_CANCEL_ASYNCHRONOUS		1


#undef	_POSIX_THREAD_PRIORITY_SCHEDULING
#define	_POSIX_THREAD_ATTR_STACKSIZE
#undef	_POSIX_THREAD_ATTR_STACKADDR
#undef	_POSIX_THREAD_PRIO_INHERIT
#undef	_POSIX_THREAD_PRIO_PROTECT


/*
 * Mutex and condition attributes are not supported.
 */
typedef void	pthread_mutex_attr_t;
typedef void	pthread_cond_attr_t;


/**
 * Thread creation attributes.
 */
typedef struct _pthread_attr {
	int     attr_state;
	long    attr_stack_addr;
	size_t  attr_stack_size;
} pthread_attr_t;


/**
 * Mutex.
 */
typedef volatile struct _pthread_mutex {
	long __attribute__((aligned(sizeof(long)))) mtx_lock;
	long __attribute__((aligned(sizeof(long)))) mtx_count;
} pthread_mutex_t;

#define	PTHREAD_MUTEX_INITIALIZER		{ 0, 0 }


/**
 * Condition.
 */
typedef volatile struct _pthread_cond {
	long __attribute__((aligned(sizeof(long)))) cond_lock;
	long __attribute__((aligned(sizeof(long)))) cond_value;
	long cond_towakeup;
} pthread_cond_t;

#define	PTHREAD_COND_INITIALIZER		{ 0, 0, 0 }


/**
 * TLS Keys.
 */
typedef volatile int	pthread_key_t;

#ifndef	PTHREAD_KEYS_MAX
#define	PTHREAD_KEYS_MAX			32
#endif


/**
 * Cleanup callback.
 */
struct _pthread_cleanup_cb {
	PLIST	list;				/* First member		*/
	void	(*cb)(void *);			/* Callback Function	*/
	void	*arg;				/* Argument		*/
};


/**
 * Thread Context.
 * @see _pthread_attr
 * @see _pthread_cleanup_cb
 */
struct _pthread {
	PLIST		 thread_list;		/* First memeber	*/

	PLOCK		 thread_lock;		/* Context lock		*/

	long		 thread_pid;		/* Process Id		*/
	long		 thread_tid;		/* Thread Id		*/
	pthread_attr_t	 thread_attr;		/* Thread Attributes	*/

	long __attribute__((aligned(sizeof(long)))) thread_state;
						/* Thread state		*/
	int		 thread_cancelstate;
	int		 thread_canceltype;

	PLIST		 thread_tlslist;	/* Thread Local Storage	*/

	int		 thread_errno;

	void		*(*thread_start_fn)(void *);
						/* Main Function	*/

	void		*thread_start_arg;	/* Main Argument	*/

	void		*thread_retval;		/* Thread's exit code	*/

	long __attribute__((aligned(sizeof(long)))) thread_finished;
						/* Wake up joined threads XXX */
	PLIST		 thread_cleanuplist;
};

typedef struct _pthread *pthread_t;


/*
 * Prototypes.
 */
PLSTATUS pthread_init_np();
void	 pthread_release_np();

int	 pthread_attr_init(pthread_attr_t *);
int	 pthread_attr_destroy(pthread_attr_t *);
int	 pthread_attr_get_np(pthread_t, pthread_attr_t *);
int	 pthread_attr_getstackaddr(const pthread_attr_t *, void **);
int	 pthread_attr_getstacksize(const pthread_attr_t *, size_t *);
int	 pthread_attr_setstacksize(pthread_attr_t *, size_t);
int	 pthread_attr_getdetachstate(pthread_attr_t *, int *);
int	 pthread_attr_setdetachstate(pthread_attr_t *, int);

int	 pthread_key_create(pthread_key_t *, void (*)(void *));
int	 pthread_key_delete(pthread_key_t);
void	*pthread_getspecific(pthread_key_t);
int	 pthread_setspecific(pthread_key_t, const void *);

int	 pthread_create(pthread_t *, pthread_attr_t *,
			void * (*)(void *), void *);

void	 pthread_exit(void *);

pthread_t pthread_self();
pthread_t pthread_lookup_np(long);

int	 pthread_mutex_init(pthread_mutex_t *, const pthread_mutex_attr_t *);
int	 pthread_mutex_destroy(pthread_mutex_t *);
int	 pthread_mutex_lock(pthread_mutex_t *);
int	 pthread_mutex_lock_np(pthread_mutex_t *, int);
int	 pthread_mutex_trylock(pthread_mutex_t *);
int	 pthread_mutex_unlock(pthread_mutex_t *);
long	 pthread_mutex_ownertid_np(pthread_mutex_t *);

int	 pthread_cond_init(pthread_cond_t *, pthread_cond_attr_t *);
int	 pthread_cond_destroy(pthread_cond_t *);
int	 pthread_cond_wait(pthread_cond_t *, pthread_mutex_t *);
int	 pthread_cond_wait_np(pthread_cond_t *, pthread_mutex_t *, long);
int	 pthread_cond_timedwait(pthread_cond_t *, pthread_mutex_t *,
				const struct timespec *);
int	 pthread_cond_signal(pthread_cond_t *);
int	 pthread_cond_signal_np(pthread_cond_t *, int, long *);
int	 pthread_cond_broadcast(pthread_cond_t *);

int	 pthread_cancel(pthread_t);

int	 pthread_setcancelstate(int, int *);
int	 pthread_setcanceltype(int, int *);
void	 pthread_cleanup_push(void(*)(void *), void *);
void	 pthread_cleanup_pop(int);
int	 pthread_cleanup_pop_np(int);
void	 pthread_testcancel();


/*
 * Unsupported.
 */
struct sched_param {};

int	 pthread_attr_setscope(pthread_attr_t *, int);
int	 pthread_setschedparam(pthread_t, int , const struct sched_param *);

#define PTHREAD_SCOPE_PROCESS		0
#define PTHREAD_SCOPE_SYSTEM		0
#define	SCHED_OTHER			0

#if 0
#define PTHREAD_EXPLICIT_SCHED
#define PTHREAD_INHERIT_SCHED
#define PTHREAD_MUTEX_DEFAULT
#define PTHREAD_MUTEX_ERRORCHECK
#define PTHREAD_MUTEX_NORMAL
#define PTHREAD_MUTEX_INITIALIZER
#define PTHREAD_MUTEX_RECURSIVE
#define PTHREAD_ONCE_INIT
#define PTHREAD_PRIO_INHERIT
#define PTHREAD_PRIO_NONE
#define PTHREAD_PRIO_PROTECT
#define PTHREAD_PROCESS_SHARED
#define PTHREAD_PROCESS_PRIVATE
#define PTHREAD_RWLOCK_INITIALIZER

int pthread_getconcurrency(void);
int pthread_getschedparam(pthread_t, int *, struct sched_param *);
int pthread_once(pthread_once_t *, void (*)(void));
int pthread_rwlock_destroy(pthread_rwlock_t *);
int pthread_rwlock_init(pthread_rwlock_t *, const pthread_rwlockattr_t *);
int pthread_rwlock_rdlock(pthread_rwlock_t *);
int pthread_rwlock_tryrdlock(pthread_rwlock_t *);
int pthread_rwlock_trywrlock(pthread_rwlock_t *);
int pthread_rwlock_unlock(pthread_rwlock_t *);
int pthread_rwlock_wrlock(pthread_rwlock_t *);
int pthread_rwlockattr_destroy(pthread_rwlockattr_t *);
int pthread_rwlockattr_getpshared(const pthread_rwlockattr_t *, int *);
int pthread_rwlockattr_init(pthread_rwlockattr_t *);
int pthread_rwlockattr_setpshared(pthread_rwlockattr_t *, int);
int pthread_setconcurrency(int);

int pthread_attr_getguardsize(const pthread_attr_t *, size_t *);
int pthread_attr_getinheritsched(const pthread_attr_t *, int *);
int pthread_attr_getschedparam(const pthread_attr_t *, struct sched_param *);
int pthread_attr_getschedpolicy(const pthread_attr_t *, int *);
int pthread_attr_getscope(const pthread_attr_t *, int *);
int pthread_attr_setguardsize(pthread_attr_t *, size_t);
int pthread_attr_setinheritsched(pthread_attr_t *, int);
int pthread_attr_setschedparam(pthread_attr_t *, const struct sched_param *);
int pthread_attr_setschedpolicy(pthread_attr_t *, int);
int pthread_attr_setstackaddr(pthread_attr_t *, void *);

int pthread_mutex_getprioceiling(const pthread_mutex_t *, int *);
int pthread_mutex_setprioceiling(pthread_mutex_t *, int, int *);
int pthread_mutexattr_destroy(pthread_mutexattr_t *);
int pthread_mutexattr_getprioceiling(const pthread_mutexattr_t *, int *);
int pthread_mutexattr_getprotocol(const pthread_mutexattr_t *, int *);
int pthread_mutexattr_getpshared(const pthread_mutexattr_t *, int *);
int pthread_mutexattr_gettype(const pthread_mutexattr_t *, int *);
int pthread_mutexattr_init(pthread_mutexattr_t *);
int pthread_mutexattr_setprioceiling(pthread_mutexattr_t *, int);
int pthread_mutexattr_setprotocol(pthread_mutexattr_t *, int);
int pthread_mutexattr_setpshared(pthread_mutexattr_t *, int);
int pthread_mutexattr_settype(pthread_mutexattr_t *, int);
#endif

#endif	/* __PTHREAD_H__ */
