/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/lib/plc/plmm.c
 * Peludo Simple Memory Manager.
 */
#include <plc.h>
#include <plspinlock.h>
#include <plist.h>


/**
 * Peludo Memory Chunk.
 * The peludo simple memory manager creates a chain of memory chunks;
 * each chunk contains at least one memory page.<br>
 * A new chunk is allocated when no more free space is available
 * in the last one.
 * <br>
 * This structure is 8 bytes aligned.
 */
typedef struct _chunk {
	PLIST		chk_list;
	unsigned long	chk_size;
	unsigned long	chk_offset;
	unsigned long	chk_count;
} CHUNK __attribute__((aligned(8)));


/**
 * Peludo Single Allocation.
 * The peludo simple memory manager creates an ALLOC structure of each
 * single allocation.
 * <br>
 * This structure is 8 bytes aligned.
 */
typedef struct _alloc {
	CHUNK		*alloc_chunk;
	unsigned long	 alloc_size;
} ALLOC __attribute__((aligned(8)));


/**
 * List of chunks.
 */
static PLOCK chunk_list_lock = PLOCK_INITIALIZED;
static PLIST chunk_list = PLIST_INITIALIZED;


/**
 * Lock the list of chunks.
 * Block all signals (including SIGCANCEL) and lock the chunk list.
 *
 * @param cset (OUT) Updated with the signal set before the lock.
 *
 * @see chunk_list_lock
 * @see _plmm_unlock()
 * @see plsigblock()
 */
static void
_plmm_lock(sigset_t *cset)
{
	plsigblock(cset);
	plock_splock(&chunk_list_lock);
}


/**
 * Unlock the list of chunks.
 * Unlock the list of chunks and set the specified signal mask.
 *
 * @param cset (IN) Signal mask to set.
 *
 * @see chunk_list_lock
 * @see _plmm_lock()
 * @see plsigunblock()
 */
static void
_plmm_unlock(sigset_t *cset)
{
	plock_unlock(&chunk_list_lock);
	plsigunblock(cset);
}


/**
 * Initialize the Peludo Simple Memory Manager.
 * This function should be called before any other plmm_xx(x) function.
 *
 * <p>Since plmm rely on the plc low level api plsys_init() should be
 * called before plmm_init(). If you are using the plc api you should
 * not care of this function because it is automatically called by
 * plinit().</p>
 *
 * @return A peludo status code is returned.
 *
 * @see PLSTATUS
 * @see plinit()
 * @see plmm_free()
 * @see plmm_malloc()
 * @see plmm_release()
 * @see plsys_init()
 */
PLSTATUS PLEXPORT
plmm_init()
{
	PLIST_INIT(&chunk_list);
	PLOCK_INIT_UNLOCKED(&chunk_list_lock);

	return PLSTATUS_SUCCESS;
}


/**
 * Release the Memory Manager.
 * Release all allocated memory chunks.
 *
 * <p>Since the plmm api rely on the plc low level api plsys_release()
 * should be called after plmm_release(). If you are using the plc api
 * you should not care of this function because it is automatically
 * called by plrelease().</p>
 *
 * @see plmm_init()
 * @see plmm_free()
 * @see plmm_malloc()
 * @see plrelease()
 * @see plsys_release()
 */
void PLEXPORT
plmm_release()
{
	sigset_t cset;
	CHUNK *chunk;

	_plmm_lock(&cset);

	while (!plist_empty(&chunk_list)) {
		chunk = PLIST_FIRST(&chunk_list, CHUNK *);
		plist_del(&chunk->chk_list);
		(void)_plsys_pfree(NULL, (long)chunk, chunk->chk_size);
	}

	_plmm_unlock(&cset);
}


/**
 * [Re]alloc a block of memory.
 *
 * @param err (OUT) If not NULL, on error it is updated with the system
 * error code (think of it as an errno value);
 *
 * @param addr (IN/OUT) Points to the address of the memory block
 * to reallocate, if that value is 0 a new block is allocated;
 * on success it is updated with the address of the new block. This
 * value is always 8 bytes aligned;
 *
 * @param size (IN) Number of bytes to [re]alloc (0 is not supported and
 * will generate an error).
 *
 * @return A Peludo Status code is returned.
 *
 * @see plmm_free()
 */
PLSTATUS PLEXPORT
plmm_realloc(long *err, long *addr, size_t size)
{
	unsigned long pages, remain, asize;
	CHUNK *chunk, *lastchunk;
	sigset_t cset;
	ALLOC *alloc;
	PLSTATUS st;

	if (addr == NULL || size == 0) {
		return PLSTATUS_INVALID;
	}

	_plmm_lock(&cset);

	/*
	 * Shrink realloc.
	 */
	if (*addr != 0) {
		alloc = ((ALLOC *)*addr) - 1;

		if (size <= alloc->alloc_size) {
			//XXX alloc->alloc_size = size;
			_plmm_unlock(&cset);
			return PLSTATUS_SUCCESS;
		}
	}

	/*
	 * Caclulate the alloc size.
	 * (There is an ALLOC structure for each single allocation).
	 */
	asize = PLALIGN(size + sizeof(ALLOC), 8);
	pages = asize & ~PLPAGE_MASK;
	remain = asize & PLPAGE_MASK;

	if (!plist_empty(&chunk_list) && pages == 0 && remain) {
		/*
		 * Try to use the last chunk free
		 * space instead of getting new pages.
		 */
		lastchunk = (CHUNK *)chunk_list.prev;

		if (remain <= lastchunk->chk_size - lastchunk->chk_offset) {

			alloc = (ALLOC *)
			    ((long)lastchunk + lastchunk->chk_offset);

			lastchunk->chk_offset += remain;
			lastchunk->chk_count++;

			chunk = lastchunk;

			goto ok;
		}
	}

	/*
	 * Get needed pages from the system.
	 * (There is a CHUNK for each _plsys_palloc() call).
	 */
	asize = PLALIGN(asize + sizeof(CHUNK), 8);

	st = _plsys_palloc(err, (long *)&chunk, asize, PLMFLAGS_R | PLMFLAGS_W);
	if (st != PLSTATUS_SUCCESS) {
		_plmm_unlock(&cset);
		return st;
	}

	PLIST_INIT(&chunk->chk_list);
	chunk->chk_size = PLALIGN(asize, PLPAGE_SIZE);
	chunk->chk_offset = asize;
	chunk->chk_count = 0;
	plist_add_tail((PLIST *)&chunk_list, (PLIST *)chunk);

	alloc = (ALLOC *)(chunk + 1);

ok:	alloc->alloc_chunk = chunk;
	alloc->alloc_size = size;

	_plmm_unlock(&cset);

	/*
	 * Do realloc.
	 */
	if (*addr != 0) {
		long oaddr, osiz;

		/*
		 * Copy old data into the new realloced address.
		 */
		oaddr = *addr;
		osiz = (((ALLOC *)oaddr) - 1)->alloc_size;
		plsys_memmove((void *)(alloc + 1), (void *)oaddr, osiz);

		plmm_free(err, oaddr);
	}

	*addr = (long)(alloc + 1);

	return PLSTATUS_SUCCESS;
}


/**
 * Release a block of memory.
 * Release a block of memory previously allocated using plmm_realloc().
 *
 * @param err (OUT) If not NULL, on error it is updated with the system
 * error code;
 *
 * @param addr (IN) Address to release.
 *
 * @see plmm_realloc()
 */
PLSTATUS PLEXPORT
plmm_free(long *err, long addr)
{
	sigset_t cset;
	CHUNK *chunk;
	ALLOC *alloc;
	PLSTATUS st;

	if (addr == 0) {
		return PLSTATUS_SUCCESS;
	}

	_plmm_lock(&cset);

	alloc = ((ALLOC *)addr) - 1;
	chunk = alloc->alloc_chunk;

	if (chunk->chk_count) {
		chunk->chk_count--;
	} else {
		plist_del(&chunk->chk_list);
		st = _plsys_pfree(err, (long)chunk, chunk->chk_size);
	}

	_plmm_unlock(&cset);

	return st;
}


