/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Peludo libc - gethostbyaddr(3), gethostbyaddr_r(3).
 */
#include <plc.h>
#include <pllog.h>

#include <netdb.h>
#include <netinet/in.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>


/**
 * Return an entry containing addresses of address family type for the host
 * with a given address.
 * Under peludo this function supportes only IPv4 addresses and returns the
 * string representation of the specified address (it does not resolve).
 * @param addr (IN) Address;
 * @param len (IN) Length of the address pointed to by <code>addr</code>;
 * @param af (IN) Address family;
 * @param he (OUT) Updated with the host entry;
 * @param buf (IN/OUT) Buffer used as a storage for the host data;
 * @param bufsize (IN) Size of buffer storage;
 * @param errorp (OUT) On error updated with the error code (h_errno code);
 * @return On success <code>he</code> is returned; on error <code>NULL</code>
 * is returned and <code>errorp</code> is updated with the error code.
 * @see http://docs.sun.com/app/docs/doc/801-6680-03/6i11qckav?a=view
 */
struct hostent PLEXPORT *
gethostbyaddr_r(const void *addr, int len, int af, struct hostent *he,
    char *buf, int buflen, int *errorp)
{
	if (af != AF_INET || len != 4) {
		pldebug("%s: af or len not supported, returning NULL.\n",
		    __FUNCTION__);
		if (errorp != NULL) {
			*errorp = HOST_NOT_FOUND;
		}

		return NULL;
	}

	if (buflen < INET_ADDRSTRLEN + 1) {
		errno = ERANGE;
		return NULL;
	}
		
	bzero(he, sizeof(*he));
	snprintf(buf, buflen, "%d.%d.%d.%d",
	    *((uint8 *)addr + 0),
	    *((uint8 *)addr + 1),
	    *((uint8 *)addr + 2),
	    *((uint8 *)addr + 3));

	he->h_name = (char *)buf;

	return he;
}


/**
 * Return an entry containing addresses of address family type for the host
 * with a given address.
 * <p>This function is not reentrant.</p>
 * Under peludo this function supports only IPv4 addresses and returns the
 * string representation of the specified address (it does not resolve).
 * @param addr (IN) Address;
 * @param len (IN) Length of the address pointed to by <code>addr</code>.
 * @return On success <code>he</code> is returned; on error <code>NULL</code>
 * is returned and <code>h_errno</code> is updated with the error code.
 * @see gethostbyaddr_r(const void *, int, int, struct hostent *, char *,
 * int, int *)
 */
struct hostent PLEXPORT *
gethostbyaddr(const void *addr, socklen_t len, int type)
{
	static char buf[INET_ADDRSTRLEN + 1];
	static struct hostent hent;
	int err;

	if (gethostbyaddr_r(addr, len, type, &hent, buf, sizeof(buf), &err)
	    == NULL) {
		h_errno = err;
		return NULL;
	}

	return &hent;
}


