/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/lib/pthread/cancel.c
 * Peludo Threads - Cancellation.
 */
#include <plc.h>
#include <pllog.h>

#include <pthread.h>
#include <signal.h>
#include <stdlib.h>

#include "pthread_p.h"


/**
 * Push a cancellation cleanup callback into the calling
 * thread's cancellation cleanup stack.
 * The cancellation cleanup callback is popped from the cancellation
 * cleanup stack and invoked when:
 * - The thread exits (that is, calls <code>pthread_exit()</code>)
 * - The thread acts upon a cancellation request
 * - The thread calls <code>pthread_cleanup_pop()</code> with a non-zero
 *   execute argument. 
 *
 * @param routine (IN) Cancellation cleanup callback;
 *
 * @param arg (IN) Argument to be passed to the cleanup callback.
 *
 * @see pthread_cancel()
 * @see pthread_cleanup_pop()
 * @see pthread_exit()
 */
void PLEXPORT
pthread_cleanup_push(void (*routine)(void *), void *arg)
{
	struct _pthread_cleanup_cb *cb;
	pthread_t self;

	if (routine != NULL) {
		cb = (struct _pthread_cleanup_cb *)malloc(sizeof(*cb));
		if (cb == NULL) {
			/* Sorry, the standard assumes it will work. */
			return;
		}

		cb->cb = routine;
		cb->arg = arg;

		self = pthread_self();

		plock_splock(&self->thread_lock);
		plist_add_head(&self->thread_cleanuplist, &cb->list);
		plock_unlock(&self->thread_lock);
	}
}


/**
 * Remove a cancellation callback from the top of the
 * calling thread's cancellation cleanup stack.
 *
 * @param execute (IN) If non-zero the cancellation callback is invoked.
 *
 * @return If a cancellation callback was removed from the stack (that is
 * the stack was not empty) a non-zero value is returned;
 * if the cancellation callback stack was empty zero is returned.
 *
 * @see pthread_cleanup_pop_np()
 * @see pthread_cleanup_push()
 */
int PLEXPORT
pthread_cleanup_pop_np(int execute)
{
	pthread_t self;
	struct _pthread_cleanup_cb *cb = NULL;

	self = pthread_self();

	plock_splock(&self->thread_lock);

	if (!plist_empty(&self->thread_cleanuplist)) {
		cb = PLIST_FIRST(&self->thread_cleanuplist,
		    struct _pthread_cleanup_cb *);
		plist_del(&cb->list);
	}

	plock_unlock(&self->thread_lock);

	if (cb != NULL) {
		if (execute) {
			cb->cb(cb->arg);
		}
		free(cb);
	}

	return (cb != NULL);
}


/**
 * Remove a cancellation callback from the top of the
 * calling thread's cancellation cleanup stack.
 *
 * @param execute (IN) If non-zero the cancellation callback is invoked.
 *
 * @see pthread_cleanup_pop_np()
 * @see pthread_cleanup_push()
 */
void PLEXPORT
pthread_cleanup_pop(int execute)
{
	(void)pthread_cleanup_pop_np(execute);
}


/**
 * Requests a thread to be canceled.
 * The target threads cancelability state and type determines when the
 * cancellation takes effect. When the cancellation is acted on,
 * the cancellation cleanup handlers associated to the specified thread
 * are called. When the last cancellation cleanup handler returns, the
 * thread-specific data destructor functions are called. When the last
 * destructor function returns, the thread is terminated.
 * <br>
 * The cancellation processing in the target thread runs asynchronously with
 * respect to the calling thread returning from <code>pthread_cancel()</code>.
 * <br>
 * Taken from: http://opengroup.org/onlinepubs/007908775/xsh/pthread_cancel.html
 *
 * @param th (IN) Thread to cancel.
 *
 * @return <code>0</code> on success; otherwise the error code.
 *
 * @see pthread_exit()
 * @see pthread_join()
 * @see pthread_setcancelstate()
 * @see pthread_cond_wait()
 * @see pthread_cond_timedwait()
 */
int PLEXPORT
pthread_cancel(pthread_t th)
{
	return pthread_kill(th, SIGCANCEL);
}


/**
 * Set calling thread's cancelability state.
 *
 * @param state (IN) Desired state: <code>PTHREAD_CANCEL_ENABLE</code>
 * or <code>PTHREAD_CANCEL_DISABLE</code>;
 *
 * @param oldstate (OUT) If not <code>NULL</code> updated with the
 * previous state.
 *
 * @return <code>0</code> on success; otherwise the error code.
 *
 * @see pthread_cancel()
 * @see pthread_setcanceltype()
 * @see pthread_testcancel()
 */
int PLEXPORT
pthread_setcancelstate(int state, int *oldstate)
{
	pthread_t self;

	if (state != PTHREAD_CANCEL_ENABLE ||
	    state != PTHREAD_CANCEL_DISABLE) {
		return EINVAL;
	}

	self = pthread_self();

	plock_splock(&self->thread_lock);

	if (oldstate != NULL) {
		*oldstate = self->thread_cancelstate;
	}

	self->thread_cancelstate = state;

	plock_unlock(&self->thread_lock);

	return 0;
}


/**
 * Set calling thread's cancelability type.
 *
 * @param state (IN) Desired type: <code>PTHREAD_CANCEL_DEFERRED</code>
 * or <code>PTHREAD_CANCEL_ASYNCHRONOUS</code>;
 *
 * @param oldtype (OUT) If not <code>NULL</code> updated with the
 * previous type.
 *
 * @return <code>0</code> on success; otherwise the error code.
 *
 * @see pthread_cancel()
 * @see pthread_setcancelstate()
 * @see pthread_testcancel()
 */
int PLEXPORT
pthread_setcanceltype(int type, int *oldtype)
{
	pthread_t self;

	if (type != PTHREAD_CANCEL_DEFERRED ||
	    type != PTHREAD_CANCEL_ASYNCHRONOUS) {
		return EINVAL;
	}

	self = pthread_self();

	plock_splock(&self->thread_lock);

	if (oldtype != NULL) {
		*oldtype = self->thread_canceltype;
	}

	self->thread_canceltype = type;

	plock_unlock(&self->thread_lock);

	return 0;
}


/**
 * Create a cancellation point in the calling thread.
 * If the cancel state is not enabled this function does nothing.
 *
 * @see pthread_exit()
 * @see pthread_setcancelstate()
 */
void PLEXPORT
pthread_testcancel()
{
	pthread_t self;

	self = pthread_self();

	plock_splock(&self->thread_lock);

	if (self->thread_cancelstate == PTHREAD_CANCEL_ENABLE &&
	    self->thread_state == PTHREAD_STATE_CANCELED) {
		plock_unlock(&self->thread_lock);
		pldebug("%s: Extiting thread from cancel point\n",
		    __FUNCTION__);
		pthread_exit(PTHREAD_CANCELED);
		/* NOTREACHED */
	}

	plock_unlock(&self->thread_lock);
}


