/*
 * Copyright 2007-2010 Claudio Castiglia
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file src/lib/plt/cond.c
 * Peludo Threads - Conditions.
 */
#include <plc.h>
#include <pllog.h>

#include <limits.h>
#include <pthread.h>

#include "pthread_p.h"


/**
 * Initialize a Condition.
 *
 * @param cond (IN) Condition to initialize.
 *
 * @return <code>0</code> is returned.
 *
 * @see pthread_cond_destroy()
 * @see pthread_cond_signal()
 * @see pthread_cond_wait()
 *
 * @todo Condition attributes.
 */
int PLEXPORT
pthread_cond_init(pthread_cond_t *cond, pthread_cond_attr_t *cattr)
{
        if (cattr != NULL) {
                pldebug("%s: Condition attributes not supported. Aborting.\n",
                    __FUNCTION__);
                plsys_exit(-1);
                /* NOTREACHED */
        }

	if (cond != NULL) {
		cond->cond_value = 0;
		cond->cond_lock = 0;
		cond->cond_towakeup = 0;
	}

	return 0;
}


/**
 * Destroy an initialized condition.
 *
 * @param cond (IN) Contidion to release.
 *
 * @return If the condition is detroyed <code>0</code> is returned;
 * if it has waiters <code>EBUSY</code> is returned;
 * on error another code is returned.
 *
 * @see pthread_cond_init()
 */
int PLEXPORT
pthread_cond_destroy(pthread_cond_t *cond)
{
//XXX	sigset_t set;
	long count;

	if (cond == NULL) {
		return EINVAL;
	}

//XXX	plsigblock(&set);
	plock_splock(&cond->cond_lock);

	count = COND_GET_COUNT(cond);

	plock_unlock(&cond->cond_lock);
//XXX	plsigunblock(&set);

#ifdef PLDEBUG
	if (count != 0) {
		pldebug("%s: Condition %p has %d waiters. Unable to destroy.\n",
		    __FUNCTION__,
		    cond,
		    count);
	}
#endif

	return ((count == 0) ? 0 : EBUSY);
}


/**
 * Wait for a condition to be signaled.
 * This function is a cacnellation point.
 *
 * @param cond (IN) Condition to wait for;
 *
 * @param mutex (IN) Pointer to an owned mutex;
 *
 * @param timeout (IN) Timeout in milliseconds (<code>-1</code> means infinite).
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_cond_broadcast()
 * @see pthread_cond_signal()
 * @see pthread_cond_signal_np()
 */
int PLEXPORT
pthread_cond_wait_np(pthread_cond_t *cond, pthread_mutex_t *mutex, long timeout)
{
	long value, ownertid;
//XXX	sigset_t set;
	PLSTATUS st;

	if (cond == NULL || mutex == NULL)  {
		pldebug("%s: NULL parameters.\n", __FUNCTION__);

		return EINVAL;
	}

	/*
	 * The current thread should own the mutex.
	 */
	ownertid = pthread_mutex_ownertid_np(mutex);
	if (ownertid != plsys_gettid()) {
		pldebug("%s: Mutex %p owned by another thread tid=%d.\n",
		    __FUNCTION__,
		    mutex,
		    ownertid);

		return EINVAL;
	}

	/*
	 * This function is a cancellation point.
	 */
	plenter_cancel_point();

	/*
	 * Add a new waiter.
	 */
//XXX	plsigblock(&set);
	plock_splock(&cond->cond_lock);

	value = COND_GET_COUNT(cond);
	value++;
	COND_SET_COUNT(cond, value);
	value = cond->cond_value;

	/*
	 * Unlocking of mutex and entering plsys_twait()
	 * should be an atomic operation.
	 * That's why the following do-while mess.
	 */
	(void)pthread_mutex_unlock(mutex);

	do {
		plock_unlock(&cond->cond_lock);
//XXX		plsigunblock(&set);

		/*
		 * Put the thread to sleep until
		 * someone calls pthread_cond_signal().
		 */
/* Too noisy */
#if 0
		pldebug("%s: Going to sleep on condition %p...\n",
		    __FUNCTION__,
		    cond);
#endif

		st = plsys_twait((void *)&cond->cond_value, value, timeout);
		switch (st) {
		case PLSTATUS_TRY_AGAIN:
			if (errno == EINTR) {
				pthread_testcancel();
			}

		case PLSTATUS_SUCCESS:
		case PLSTATUS_TIMEOUT:
//XXX			plsigblock(&set);
			plock_splock(&cond->cond_lock);

			/*
			 * Check whether the condition was signaled.
			 */
			if (COND_IS_SIGNALED(cond)) {
				/*
				 * Wake up if we should.
				 */
				if (cond->cond_towakeup) {
					cond->cond_towakeup--;
				}

				st = PLSTATUS_SUCCESS;
				break;
			}

			/*
 			 * No wakeup, check timeout.
 			 */
			if (st == PLSTATUS_TIMEOUT) {
				break;
			}

			/*
			 * No wakeup, sleep again.
			 */
			value = cond->cond_value;
			break;

		default:
			/*
			 * Something unexpected happened.
			 */
			pldebug("%s: Sleep into condition %p failed, "
			    "peludo status=<%s>, errno=%d. Abort.\n",
			    __FUNCTION__,
			    cond,
			    plstrstatus(st),
			    errno);
	
			plsys_exit(-1);
			/* NOTREACHED */
		}

	} while (st != PLSTATUS_SUCCESS && st != PLSTATUS_TIMEOUT);

	/*
	 * We are no more a waiter.
	 */
	value = COND_GET_COUNT(cond);
	value--;
	COND_SET_COUNT(cond, value);

 	/*
	 * Clear the signaled flag when the are no more waiters.
	 */
	if (value == 0) {
		COND_CLEAR_FLAG(cond, COND_FLAG_SIGNALED);
	}

	plock_unlock(&cond->cond_lock);
//XXX	plsigunblock(&set);

	/*
	 * This function is a cancellation point.
	 */
	plleave_cancel_point();

	/*
	 * Own the mutex and return.
	 */
	return ((st == PLSTATUS_SUCCESS) ?
	    pthread_mutex_lock(mutex) : ETIMEDOUT);
}


/**
 * Block on a condition until it is signaled.
 * This function is a cancellation point.
 *
 * @param cond (IN) Condition to wait for;
 *
 * @param mutex (IN) Pointer to an owned mutex;
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_cond_broadcast()
 * @see pthread_cond_signal()
 * @see pthread_cond_signal_np()
 * @see pthread_cond_wait_np()
 */
int PLEXPORT
pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
	return pthread_cond_wait_np(cond, mutex, -1);
}


/**
 * Block on a condition until it is signaled or a timeout expires.
 * This function is a cancellation point.
 *
 * @param cond (IN) Condition to wait for;
 *
 * @param mutex (IN) Pointer to an owned mutex;
 *
 * @param at (IN) Timeout.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_cond_broadcast()
 * @see pthread_cond_signal()
 * @see pthread_cond_signal_np()
 * @see pthread_cond_wait_np()
 */
int PLEXPORT
pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex,
    const struct timespec *at)
{
	struct timeval tv;
	long ms;

	(void)plsys_gettimeofday(&tv, NULL);

	ms = (at->tv_nsec / 1000000 + at->tv_sec * 1000);
	ms -= (tv.tv_usec / 1000 + tv.tv_sec * 1000);

	if (ms > INT_MAX || ms < 0) {
		return EINVAL;
	}

	return pthread_cond_wait_np(cond, mutex, ms);
}


/**
 * Wakeup one or all threads blocked on a condition.
 *
 * @param cond (IN) Condition;
 *
 * @param all (IN) A non-zero value wakes all blocked threads
 * otherwise only one thread is awaken (if there is any blocked thread);
 *
 * @param pwoken (OUT) If not <code>NULL</code> it is uptaded with the
 * number of woken threads.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_cond_wait_np()
 */
int PLEXPORT
pthread_cond_signal_np(pthread_cond_t *cond, int all, long *pwoken)
{
	long value, woken;
//XXX	sigset_t set;

	if (cond == NULL) {
		pldebug("%s: NULL parameter.\n", __FUNCTION__);
		return EINVAL;
	}

//XXX	plsigblock(&set);
	plock_splock(&cond->cond_lock);

	pldebug("%s: Signaling waiters of %p.\n",
	    __FUNCTION__,
	    cond);

	value = COND_GET_COUNT(cond);

	if (value != 0) {
		cond->cond_towakeup = ((all) ? value : 1);
		COND_SET_FLAG(cond, COND_FLAG_SIGNALED);

		woken = plsys_tsignal((void *)&cond->cond_value, all);

		pldebug("%s: Threads to wakeup=%d. currently sleeping=%d\n",
		    __FUNCTION__,
		    cond->cond_towakeup,
		    woken);

		if (woken < 0) {
			pldebug("%s: Failed, errno=%d. Aborting.\n",
			    __FUNCTION__,
			    errno);

			plsys_exit(-1);
			/* NOTREACHED */
		}
	}

	if (pwoken != NULL) {
		*pwoken = value;
	}

	plock_unlock(&cond->cond_lock);
//XXX	plsigunblock(&set);

	return PLSTATUS_SUCCESS;
}


/**
 * Wakeup at least one thread blocked on a condition
 * (if any threads are blocked).
 *
 * @param cond (IN) Condition;
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_cond_broadcast()
 * @see pthread_cond_signal_np()
 * @see pthread_cond_wait_np()
 */
int PLEXPORT
pthread_cond_signal(pthread_cond_t *cond)
{
	return pthread_cond_signal_np(cond, 0, NULL);
}


/**
 * Wakeup all threads blocked on a condition.
 *
 * @param cond (IN) Condition;
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_cond_signal()
 * @see pthread_cond_signal_np()
 * @see pthread_cond_wait_np()
 */
int PLEXPORT
pthread_cond_broadcast(pthread_cond_t *cond)
{
	return pthread_cond_signal_np(cond, 1, NULL);
}


